from datetime import datetime
from typing import Any, Dict, List, Tuple

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import MongoDBStore, Store


class Analytic:
    def __init__(self, id: str, client_id: str, type: str, execution_duration: int, model_id: str, committed_at: datetime):
        self.id = id
        self.client_id = client_id
        self.type = type
        self.execution_duration = execution_duration
        self.model_id = model_id
        self.committed_at = committed_at


class AnalyticStore(Store[Analytic]):
    pass


def _validate_analytic(analytic: dict) -> Tuple[bool, str]:
    if "client_id" not in analytic:
        return False, "client_id is required"
    if "type" not in analytic or analytic["type"] not in ["training", "inference"]:
        return False, "type must be either 'training' or 'inference'"
    return analytic, ""


def _complete_analytic(analytic: dict) -> dict:
    if "committed_at" not in analytic:
        analytic["committed_at"] = datetime.now()


class MongoDBAnalyticStore(AnalyticStore, MongoDBStore[Analytic]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)
        self.database[self.collection].create_index([("client_id", pymongo.DESCENDING)])

    def get(self, id: str) -> Analytic:
        return super().get(id)

    def update(self, id: str, item: Analytic) -> Tuple[bool, Any]:
        pass

    def add(self, item: Analytic) -> Tuple[bool, Any]:
        valid, msg = _validate_analytic(item)
        if not valid:
            return False, msg

        _complete_analytic(item)

        return super().add(item)

    def delete(self, id: str) -> bool:
        pass

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs) -> Dict[int, List[Analytic]]:
        return super().list(limit, skip, sort_key or "committed_at", sort_order, **kwargs)

    def count(self, **kwargs) -> int:
        return super().count(**kwargs)




import sys
import pytest

from fedn.tests.stores.helpers.database_helper import mongo_connection, sql_connection, postgres_connection


# These lines ensure that pytests trigger breakpoints when assertions fail during debugging
def is_debugging():
    return 'debugpy' in sys.modules
    
# enable_stop_on_exceptions if the debugger is running during a test
if is_debugging():
  @pytest.hookimpl(tryfirst=True)
  def pytest_exception_interact(call):
    raise call.excinfo.value
    
  @pytest.hookimpl(tryfirst=True)
  def pytest_internalerror(excinfo):
    raise excinfo.value


import pytest
import pymongo

import datetime
import uuid
import itertools

from fedn.network.storage.dbconnection import DatabaseConnection

@pytest.fixture
def test_statuses():
    start_time = datetime.datetime(2021, 1, 4, 1, 2, 4)
   
    return [{"log_level":"test_log_level1", "status":"status1", "timestamp":start_time+datetime.timedelta(days=2), "type":"test_type"},
            {"log_level":"test_log_level2", "status":"status2", "timestamp":start_time+datetime.timedelta(days=4), "type":"test_type"},
            {"log_level":"test_log_level3", "status":"status3", "timestamp":start_time+datetime.timedelta(days=5), "type":"test_type"},
            {"log_level":"test_log_level4", "status":"status4", "timestamp":start_time+datetime.timedelta(days=1), "type":"test_type2"},
            {"log_level":"test_log_level5", "status":"status5", "timestamp":start_time+datetime.timedelta(days=6), "type":"test_type2"},
            {"log_level":"test_log_level6", "status":"status6", "timestamp":start_time+datetime.timedelta(days=3), "type":"test_type2"},
            ]

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_statuses):
    for c in test_statuses:
        res, _ = mongo_connection.status_store.add(c)
        assert res == True

    for c in test_statuses:
        res, _ = postgres_connection.status_store.add(c)
        assert res == True

    for c in test_statuses:
        res, _ = sql_connection.status_store.add(c)
        assert res == True

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    # TODO:Clean up



@pytest.fixture
def options():
    sorting_keys = (#None, 
                        "log_level",
                        "timestamp",
                        #"invalid_key"
                        ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"log_level":"test_log_level6"}, {"type":"test_type2"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))

class TestStatusStore:

    def test_add_update_delete(self, postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection):
        pass

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                res = db_1.status_store.list(*opt, **kwargs)
                count, gathered_statuses = res["count"], res["result"]

                res = db_2.status_store.list(*opt, **kwargs)
                count2, gathered_statuses2 = res["count"], res["result"]
                #TODO: The count is not equal to the number of clients in the list, but the number of clients returned by the query before skip and limit
                #It is not clear what is the intended behavior
                # assert(count == len(gathered_clients))
                # assert count == count2
                assert len(gathered_statuses) == len(gathered_statuses2)

                for i in range(len(gathered_statuses)):
                    assert gathered_statuses2[i]["log_level"] == gathered_statuses[i]["log_level"]
                    # TODO: Status is saved differently in the two databases, why?
                    # assert gathered_statuses2[i]["timestamp"] == gathered_statuses[i]["timestamp"]
                    assert gathered_statuses2[i]["status"] == gathered_statuses[i]["status"]





import pytest
import pymongo

import time
import datetime
import uuid
import itertools

from fedn.network.storage.dbconnection import DatabaseConnection


@pytest.fixture
def test_sessions():
    start_time = datetime.datetime(2021, 1, 4, 1, 2, 4)
    
    model = {"id":str(uuid.uuid4()), "key":"models", "model":str(uuid.uuid4()), "parent_model":"test_parent_model", "session_id":None, "committed_at":start_time, "name":"test_name1", "active":True}

    session_config = {"aggregator":"test_aggregator", "round_timeout":100, "buffer_size":100, "delete_models_storage":True, 
                      "clients_required":10, "validate":True, "helper_type":"test_helper_type", "model_id":model["model"]}

    return model,[{"id":str(uuid.uuid4()),"name":"sessionname1",  "session_config":session_config},
            {"id":str(uuid.uuid4()),"name":"sessionname3",  "session_config":session_config},
            {"id":str(uuid.uuid4()),"name":"sessionname5",  "session_config":session_config},
            {"id":str(uuid.uuid4()),"name":"sessionname4",  "session_config":session_config},
            {"id":str(uuid.uuid4()),"name":"sessionname6",  "session_config":session_config},
            {"id":str(uuid.uuid4()),"name":"sessionname2",   "session_config":session_config},
            ]

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_sessions):
    model, test_sessions = test_sessions
    res, _ = mongo_connection.model_store.add(model)
    assert res == True
    res, _ = postgres_connection.model_store.add(model)
    assert res == True
    res, _ = sql_connection.model_store.add(model)
    assert res == True

    for c in test_sessions:
        res, _ = mongo_connection.session_store.add(c)
        assert res == True

    for c in test_sessions:
        res, _ = postgres_connection.session_store.add(c)
        assert res == True

    for c in test_sessions:
        res, _ = sql_connection.session_store.add(c)
        assert res == True

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    # TODO:Clean up



@pytest.fixture
def options():
    sorting_keys = (#None, 
                    "name",
                    #"committed_at",
                    #"invalid_key"
                    ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"status":"test_status4"}, {"status":"blah"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))


class TestSessionStore:

    def test_add_update_delete(self, postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection):
        pass


    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:
                res = db_1.session_store.list(*opt, **kwargs)
                count, gathered_sessions = res["count"], res["result"]

                res = db_2.session_store.list(*opt, **kwargs)
                count2, gathered_sessions2 = res["count"], res["result"]
                #TODO: The count is not equal to the number of clients in the list, but the number of clients returned by the query before skip and limit
                #It is not clear what is the intended behavior
                # assert(count == len(gathered_clients))
                # assert count == count2
                assert len(gathered_sessions) == len(gathered_sessions2)

                for i in range(len(gathered_sessions)):
                    #assert gathered_sessions2[i]["id"] == gathered_sessions[i]["id"]
                    assert gathered_sessions2[i]["name"] == gathered_sessions[i]["name"]





import uuid
from abc import abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

import pymongo
from bson import ObjectId
from pymongo.database import Database
from sqlalchemy import String, func, select
from sqlalchemy.orm import Mapped, mapped_column
from werkzeug.utils import secure_filename

from fedn.network.storage.statestore.stores.store import MongoDBStore, MyAbstractBase, SQLStore, Store


def from_document(data: dict, active_package: dict):
    active = False
    if active_package:
        if "id" in active_package and "id" in data:
            active = active_package["id"] == data["id"]

    return {
        "id": data["id"] if "id" in data else None,
        "key": data["key"] if "key" in data else None,
        "committed_at": data["committed_at"] if "committed_at" in data else None,
        "description": data["description"] if "description" in data else None,
        "file_name": data["file_name"] if "file_name" in data else None,
        "helper": data["helper"] if "helper" in data else None,
        "name": data["name"] if "name" in data else None,
        "storage_file_name": data["storage_file_name"] if "storage_file_name" in data else None,
        "active": active,
    }


class Package:
    def __init__(
        self,
        id: str,
        key: str,
        committed_at: datetime,
        description: str,
        file_name: str,
        helper: str,
        name: str,
        storage_file_name: str,
        active: bool = False,
    ):
        self.key = key
        self.committed_at = committed_at
        self.description = description
        self.file_name = file_name
        self.helper = helper
        self.id = id
        self.name = name
        self.storage_file_name = storage_file_name
        self.active = active


class PackageStore(Store[Package]):
    @abstractmethod
    def set_active(self, id: str) -> bool:
        pass

    @abstractmethod
    def get_active(self) -> Package:
        pass

    @abstractmethod
    def set_active_helper(self, helper: str) -> bool:
        pass

    @abstractmethod
    def delete_active(self):
        pass


def allowed_file_extension(filename: str, ALLOWED_EXTENSIONS={"gz", "bz2", "tar", "zip", "tgz"}) -> bool:
    """Check if file extension is allowed.

    :param filename: The filename to check.
    :type filename: str
    :return: True and extension str if file extension is allowed, else False and None.
    :rtype: Tuple (bool, str)
    """
    if "." in filename:
        extension = filename.rsplit(".", 1)[1].lower()
        if extension in ALLOWED_EXTENSIONS:
            return True

    return False


def validate(item: Package) -> Tuple[bool, str]:
    if "file_name" not in item or not item["file_name"]:
        return False, "File name is required"

    if not allowed_file_extension(item["file_name"]):
        return False, "File extension not allowed"

    if "helper" not in item or not item["helper"]:
        return False, "Helper is required"

    return True, ""


class MongoDBPackageStore(PackageStore, MongoDBStore[Package]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)
        self.database[self.collection].create_index([("id", pymongo.DESCENDING)])

    def get(self, id: str) -> Package:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the docuemnt _id
        return: The entity
        """
        document = self.database[self.collection].find_one({"id": id})

        if document is None:
            return None

        response_active = self.database[self.collection].find_one({"key": "active"})

        return from_document(document, response_active)

    def _complement(self, item: Package):
        if "id" not in item or item["id"] is None:
            item["id"] = str(uuid.uuid4())

        if "key" not in item or item["key"] is None:
            item["key"] = "package_trail"

        if "committed_at" not in item or item["committed_at"] is None:
            item["committed_at"] = datetime.now()

        extension = item["file_name"].rsplit(".", 1)[1].lower()

        if "storage_file_name" not in item or item["storage_file_name"] is None:
            storage_file_name = secure_filename(f"{str(uuid.uuid4())}.{extension}")
            item["storage_file_name"] = storage_file_name

    def set_active(self, id: str) -> bool:
        """Set the active entity
        param id: The id of the entity
            type: str
        return: Whether the operation was successful
        """
        kwargs = {"_id": ObjectId(id)} if ObjectId.is_valid(id) else {"id": id}
        kwargs["key"] = "package_trail"

        document = self.database[self.collection].find_one(kwargs)

        if document is None:
            return False

        committed_at = datetime.now()
        obj_to_insert = {
            "key": "active",
            "id": document["id"] if "id" in document else "",
            "committed_at": committed_at,
            "description": document["description"] if "description" in document else "",
            "file_name": document["file_name"] if "file_name" in document else "",
            "helper": document["helper"] if "helper" in document else "",
            "name": document["name"] if "name" in document else "",
            "storage_file_name": document["storage_file_name"],
        }

        self.database[self.collection].update_one({"key": "active"}, {"$set": obj_to_insert}, upsert=True)

        return True

    def get_active(self) -> Package:
        """Get the active entity
        return: The entity
        """
        kwargs = {"key": "active"}
        response = self.database[self.collection].find_one(kwargs)
        if response is None:
            return None

        active_package = {"id": response["id"]} if "id" in response else {}

        return from_document(response, active_package=active_package)

    def set_active_helper(self, helper: str) -> bool:
        """Set the active helper
        param helper: The helper to set as active
            type: str
        return: Whether the operation was successful
        """
        if not helper or helper == "" or helper not in ["numpyhelper", "binaryhelper", "androidhelper"]:
            raise ValueError()

        try:
            self.database[self.collection].update_one({"key": "active"}, {"$set": {"helper": helper}}, upsert=True)
        except Exception:
            return False

    def update(self, id: str, item: Package) -> bool:
        raise NotImplementedError("Update not implemented for PackageStore")

    def add(self, item: Package) -> Tuple[bool, Any]:
        valid, message = validate(item)
        if not valid:
            return False, message

        self._complement(item)

        return super().add(item)

    def delete(self, id: str) -> bool:
        kwargs = {"_id": ObjectId(id)} if ObjectId.is_valid(id) else {"id": id}
        kwargs["key"] = "package_trail"
        document = self.database[self.collection].find_one(kwargs)

        if document is None:
            return False

        result = super().delete(document["_id"])

        if not result:
            return False

        kwargs["key"] = "active"

        document_active = self.database[self.collection].find_one(kwargs)

        if document_active is not None:
            return super().delete(document_active["_id"])

        return True

    def delete_active(self) -> bool:
        kwargs = {"key": "active"}

        document_active = self.database[self.collection].find_one(kwargs)

        if document_active is None:
            return False

        return super().delete(document_active["_id"])

    def list(
        self,
        limit: int,
        skip: int,
        sort_key: str,
        sort_order=pymongo.DESCENDING,
        **kwargs,
    ) -> Dict[int, List[Package]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        kwargs["key"] = "package_trail"

        response = self.database[self.collection].find(kwargs).sort(sort_key or "committed_at", sort_order).skip(skip or 0).limit(limit or 0)

        count = self.database[self.collection].count_documents(kwargs)

        response_active = self.database[self.collection].find_one({"key": "active"})

        result = [from_document(item, response_active) for item in response]

        return {"count": count, "result": result}

    def count(self, **kwargs) -> int:
        kwargs["key"] = "package_trail"
        return super().count(**kwargs)


class PackageModel(MyAbstractBase):
    __tablename__ = "packages"

    active: Mapped[bool] = mapped_column(default=False)
    description: Mapped[Optional[str]] = mapped_column(String(255))
    file_name: Mapped[str] = mapped_column(String(255))
    helper: Mapped[str] = mapped_column(String(255))
    name: Mapped[str] = mapped_column(String(255))
    storage_file_name: Mapped[str] = mapped_column(String(255))


def from_row(row: PackageModel) -> Package:
    return {
        "id": row.id,
        "committed_at": row.committed_at,
        "description": row.description,
        "file_name": row.file_name,
        "helper": row.helper,
        "name": row.name,
        "storage_file_name": row.storage_file_name,
        "active": row.active,
    }


class SQLPackageStore(PackageStore, SQLStore[Package]):
    def __init__(self, Session):
        super().__init__(Session)

    def _complement(self, item: Package):
        # TODO: Not complemented the same way as in MongoDBStore
        if "committed_at" not in item or item["committed_at"] is None:
            item["committed_at"] = datetime.now()

        extension = item["file_name"].rsplit(".", 1)[1].lower()

        if "storage_file_name" not in item or item["storage_file_name"] is None:
            storage_file_name = secure_filename(f"{str(uuid.uuid4())}.{extension}")
            item["storage_file_name"] = storage_file_name

    def add(self, item: Package) -> Tuple[bool, Any]:
        valid, message = validate(item)
        if not valid:
            return False, message

        self._complement(item)
        with self.Session() as session:
            item = PackageModel(
                committed_at=item["committed_at"],
                description=item["description"] if "description" in item else "",
                file_name=item["file_name"],
                helper=item["helper"],
                name=item["name"] if "name" in item else "",
                storage_file_name=item["storage_file_name"],
            )
            session.add(item)
            session.commit()
            return True, from_row(item)

    def get(self, id: str) -> Package:
        with self.Session() as session:
            stmt = select(PackageModel).where(PackageModel.id == id)
            item = session.scalars(stmt).first()
            if item is None:
                return None
            return from_row(item)

    def update(self, id: str, item: Package) -> bool:
        raise NotImplementedError

    def delete(self, id: str) -> bool:
        with self.Session() as session:
            stmt = select(PackageModel).where(PackageModel.id == id)
            item = session.scalars(stmt).first()
            if item is None:
                return False
            session.delete(item)
            session.commit()
            return True

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs):
        with self.Session() as session:
            stmt = select(PackageModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(PackageModel, key) == value)

            _sort_order: str = "DESC" if sort_order == pymongo.DESCENDING else "ASC"
            _sort_key: str = sort_key or "committed_at"

            if _sort_key in PackageModel.__table__.columns:
                sort_obj = PackageModel.__table__.columns.get(_sort_key) if _sort_order == "ASC" else PackageModel.__table__.columns.get(_sort_key).desc()

                stmt = stmt.order_by(sort_obj)

            if limit:
                stmt = stmt.offset(skip or 0).limit(limit)
            elif skip:
                stmt = stmt.offset(skip)

            items = session.scalars(stmt).all()

            result = []
            for i in items:
                result.append(from_row(i))

            count = session.scalar(select(func.count()).select_from(PackageModel))

            return {"count": count, "result": result}

    def count(self, **kwargs):
        with self.Session() as session:
            stmt = select(func.count()).select_from(PackageModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(PackageModel, key) == value)

            count = session.scalar(stmt)

            return count

    def set_active(self, id: str):
        with self.Session() as session:
            active_stmt = select(PackageModel).where(PackageModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                active_item.active = False

            stmt = select(PackageModel).where(PackageModel.id == id)
            item = session.scalars(stmt).first()

            if item is None:
                return None

            item.active = True
            session.commit()
        return True

    def get_active(self) -> Package:
        with self.Session() as session:
            active_stmt = select(PackageModel).where(PackageModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                return from_row(active_item)
            return None

    def set_active_helper(self, helper: str) -> bool:
        if not helper or helper == "" or helper not in ["numpyhelper", "binaryhelper", "androidhelper"]:
            raise ValueError()

        with self.Session() as session:
            active_stmt = select(PackageModel).where(PackageModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                active_item.helper = helper
                session.commit()
                return True
            item = PackageModel(
                committed_at=datetime.now(),
                description="",
                file_name="",
                helper=helper,
                name="",
                storage_file_name="",
                active=True,
            )

            session.add(item)
            session.commit()

    def delete_active(self) -> bool:
        with self.Session() as session:
            active_stmt = select(PackageModel).where(PackageModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                active_item.active = False
                session.commit()
                return True
            return False



from flask import Blueprint, jsonify, request

from fedn.common.log_config import logger
from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.shared import package_store
from fedn.network.api.v1.shared import api_version

bp = Blueprint("helper", __name__, url_prefix=f"/api/{api_version}/helpers")


@bp.route("/active", methods=["GET"])
@jwt_auth_required(role="admin")
def get_active_helper():
    """Get active helper
    Retrieves the active helper
    ---
    tags:
        - Helpers
    responses:
        200:
            description: Active helper
        404:
            description: No active helper
        500:
            description: An unexpected error occurred
    """
    try:
        active_package = package_store.get_active()
        if active_package is None:
            return jsonify({"message": "No active helper"}), 404

        response = active_package["helper"]

        return jsonify(response), 200
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return jsonify({"message": "An unexpected error occurred"}), 500


@bp.route("/active", methods=["PUT"])
@jwt_auth_required(role="admin")
def set_active_helper():
    """Set active helper
    Sets the active helper
    ---
    tags:
        - Helpers
    responses:
        200:
            description: Active helper set
        500:
            description: An unexpected error occurred
    """
    try:
        data = request.get_json()
        helper = data["helper"]
        package_store.set_active_helper(helper)

        return jsonify({"message": "Active helper set"}), 200
    except ValueError:
        return jsonify({"message": "Helper is required to be either 'numpyhelper', 'binaryhelper' or 'androidhelper'"}), 400
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
        return jsonify({"message": "An unexpected error occurred"}), 500



from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

import pymongo
from bson import ObjectId
from pymongo.database import Database
from sqlalchemy import String, func, or_, select
from sqlalchemy.orm import Mapped, mapped_column

from fedn.network.storage.statestore.stores.store import MongoDBStore, MyAbstractBase, SQLStore, Store

from .shared import from_document


class Combiner:
    def __init__(
        self,
        id: str,
        name: str,
        address: str,
        certificate: str,
        config: dict,
        fqdn: str,
        ip: str,
        key: str,
        parent: dict,
        port: int,
        status: str,
        updated_at: str,
    ):
        self.id = id
        self.name = name
        self.address = address
        self.certificate = certificate
        self.config = config
        self.fqdn = fqdn
        self.ip = ip
        self.key = key
        self.parent = parent
        self.port = port
        self.status = status
        self.updated_at = updated_at


class CombinerStore(Store[Combiner]):
    pass


class MongoDBCombinerStore(MongoDBStore[Combiner]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str) -> Combiner:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the name (property)
        return: The entity
        """
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            document = self.database[self.collection].find_one({"_id": id_obj})
        else:
            document = self.database[self.collection].find_one({"name": id})

        if document is None:
            return None

        return from_document(document)

    def update(self, id: str, item: Combiner) -> bool:
        raise NotImplementedError("Update not implemented for CombinerStore")

    def add(self, item: Combiner) -> Tuple[bool, Any]:
        return super().add(item)

    def delete(self, id: str) -> bool:
        if ObjectId.is_valid(id):
            kwargs = {"_id": ObjectId(id)}
        else:
            return False

        document = self.database[self.collection].find_one(kwargs)

        if document is None:
            return False

        return super().delete(document["_id"])

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs) -> Dict[int, List[Combiner]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        response = super().list(limit, skip, sort_key or "updated_at", sort_order, **kwargs)

        return response

    def count(self, **kwargs) -> int:
        return super().count(**kwargs)


class CombinerModel(MyAbstractBase):
    __tablename__ = "combiners"

    address: Mapped[str] = mapped_column(String(255))
    fqdn: Mapped[Optional[str]] = mapped_column(String(255))
    ip: Mapped[Optional[str]] = mapped_column(String(255))
    name: Mapped[str] = mapped_column(String(255))
    parent: Mapped[Optional[str]] = mapped_column(String(255))
    port: Mapped[int]
    updated_at: Mapped[datetime] = mapped_column(default=datetime.now())


def from_row(row: CombinerModel) -> Combiner:
    return {
        "id": row.id,
        "committed_at": row.committed_at,
        "address": row.address,
        "ip": row.ip,
        "name": row.name,
        "parent": row.parent,
        "fqdn": row.fqdn,
        "port": row.port,
        "updated_at": row.updated_at,
    }


class SQLCombinerStore(CombinerStore, SQLStore[Combiner]):
    def __init__(self, Session):
        super().__init__(Session)

    def get(self, id: str) -> Combiner:
        with self.Session() as session:
            stmt = select(CombinerModel).where(or_(CombinerModel.id == id, CombinerModel.name == id))
            item = session.scalars(stmt).first()
            if item is None:
                return None
            return from_row(item)

    def update(self, id, item):
        raise NotImplementedError

    def add(self, item):
        with self.Session() as session:
            entity = CombinerModel(
                address=item["address"],
                fqdn=item["fqdn"],
                ip=item["ip"],
                name=item["name"],
                parent=item["parent"],
                port=item["port"],
            )
            session.add(entity)
            session.commit()
            return True, from_row(entity)

    def delete(self, id: str) -> bool:
        with self.Session() as session:
            stmt = select(CombinerModel).where(CombinerModel.id == id)
            item = session.scalars(stmt).first()
            if item is None:
                return False
            session.delete(item)
            return True

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs):
        with self.Session() as session:
            stmt = select(CombinerModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(CombinerModel, key) == value)

            _sort_order: str = "DESC" if sort_order == pymongo.DESCENDING else "ASC"
            _sort_key: str = sort_key or "committed_at"

            if _sort_key in CombinerModel.__table__.columns:
                sort_obj = CombinerModel.__table__.columns.get(_sort_key) if _sort_order == "ASC" else CombinerModel.__table__.columns.get(_sort_key).desc()

                stmt = stmt.order_by(sort_obj)

            if limit:
                stmt = stmt.offset(skip or 0).limit(limit)
            elif skip:
                stmt = stmt.offset(skip)

            items = session.scalars(stmt).all()

            result = []
            for i in items:
                result.append(from_row(i))

            count = session.scalar(select(func.count()).select_from(CombinerModel))

            return {"count": count, "result": result}

    def count(self, **kwargs):
        with self.Session() as session:
            stmt = select(func.count()).select_from(CombinerModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(CombinerModel, key) == value)

            count = session.scalar(stmt)

            return count



"""FednClient class for interacting with the FEDn network."""

import enum
import json
import os
import threading
import time
import uuid
from io import BytesIO
from typing import Any, Optional, Tuple, Union

import requests

import fedn.network.grpc.fedn_pb2 as fedn
from fedn.common.config import FEDN_AUTH_SCHEME, FEDN_PACKAGE_EXTRACT_DIR
from fedn.common.log_config import logger
from fedn.network.clients.grpc_handler import GrpcHandler
from fedn.network.clients.package_runtime import PackageRuntime
from fedn.utils.dispatcher import Dispatcher

# Constants for HTTP status codes
HTTP_STATUS_OK = 200
HTTP_STATUS_NO_CONTENT = 204
HTTP_STATUS_BAD_REQUEST = 400
HTTP_STATUS_UNAUTHORIZED = 401
HTTP_STATUS_NOT_FOUND = 404
HTTP_STATUS_PACKAGE_MISSING = 203

# Default timeout for requests
REQUEST_TIMEOUT = 10  # seconds


class GrpcConnectionOptions:
    """Options for configuring the GRPC connection."""

    def __init__(self, status: str, host: str, fqdn: str, package: str, ip: str, port: int, helper_type: str) -> None:
        """Initialize GrpcConnectionOptions."""
        self.status = status
        self.host = host
        self.fqdn = fqdn
        self.package = package
        self.ip = ip
        self.port = port
        self.helper_type = helper_type

    @classmethod
    def from_dict(cls, config: dict) -> "GrpcConnectionOptions":
        """Create a GrpcConnectionOptions instance from a JSON string."""
        return cls(
            status=config.get("status", ""),
            host=config.get("host", ""),
            fqdn=config.get("fqdn", ""),
            package=config.get("package", ""),
            ip=config.get("ip", ""),
            port=config.get("port", 0),
            helper_type=config.get("helper_type", ""),
        )


class ConnectToApiResult(enum.Enum):
    """Enum for representing the result of connecting to the FEDn API."""

    Assigned = 0
    ComputePackageMissing = 1
    UnAuthorized = 2
    UnMatchedConfig = 3
    IncorrectUrl = 4
    UnknownError = 5


def get_compute_package_dir_path() -> str:
    """Get the directory path for the compute package."""
    if FEDN_PACKAGE_EXTRACT_DIR:
        result = os.path.join(os.getcwd(), FEDN_PACKAGE_EXTRACT_DIR)
    else:
        dirname = "compute-package-" + time.strftime("%Y%m%d-%H%M%S")
        result = os.path.join(os.getcwd(), dirname)

    if not os.path.exists(result):
        os.mkdir(result)

    return result


class FednClient:
    """Client for interacting with the FEDn network."""

    def __init__(
        self, train_callback: Optional[callable] = None, validate_callback: Optional[callable] = None, predict_callback: Optional[callable] = None
    ) -> None:
        """Initialize the FednClient."""
        self.train_callback = train_callback
        self.validate_callback = validate_callback
        self.predict_callback = predict_callback

        path = get_compute_package_dir_path()
        self._package_runtime = PackageRuntime(path)

        self.dispatcher: Optional[Dispatcher] = None
        self.grpc_handler: Optional[GrpcHandler] = None

    def set_train_callback(self, callback: callable) -> None:
        """Set the train callback."""
        self.train_callback = callback

    def set_validate_callback(self, callback: callable) -> None:
        """Set the validate callback."""
        self.validate_callback = callback

    def set_predict_callback(self, callback: callable) -> None:
        """Set the predict callback."""
        self.predict_callback = callback

    def connect_to_api(self, url: str, token: str, json: dict) -> Tuple[ConnectToApiResult, Any]:
        """Connect to the FEDn API."""
        url_endpoint = f"{url}api/v1/clients/add"
        logger.info(f"Connecting to API endpoint: {url_endpoint}")

        try:
            response = requests.post(
                url=url_endpoint,
                json=json,
                allow_redirects=True,
                headers={"Authorization": f"{FEDN_AUTH_SCHEME} {token}"},
                timeout=REQUEST_TIMEOUT,
            )

            if response.status_code == HTTP_STATUS_OK:
                logger.info("Connect to FEDn Api - Client assigned to controller")
                json_response = response.json()
                combiner_config = GrpcConnectionOptions.from_dict(json_response)
                return ConnectToApiResult.Assigned, combiner_config

            if response.status_code == HTTP_STATUS_PACKAGE_MISSING:
                json_response = response.json()
                logger.info("Connect to FEDn Api - Remote compute package missing.")
                return ConnectToApiResult.ComputePackageMissing, json_response

            if response.status_code == HTTP_STATUS_UNAUTHORIZED:
                logger.warning("Connect to FEDn Api - Unauthorized")
                return ConnectToApiResult.UnAuthorized, "Unauthorized"

            if response.status_code == HTTP_STATUS_BAD_REQUEST:
                json_response = response.json()
                msg = json_response["message"]
                logger.warning(f"Connect to FEDn Api - {msg}")
                return ConnectToApiResult.UnMatchedConfig, msg

            if response.status_code == HTTP_STATUS_NOT_FOUND:
                logger.warning("Connect to FEDn Api - Incorrect URL")
                return ConnectToApiResult.IncorrectUrl, "Incorrect URL"

        except Exception as e:
            logger.warning(f"Connect to FEDn Api - Error occurred: {str(e)}")
            return ConnectToApiResult.UnknownError, str(e)

    def download_compute_package(self, url: str, token: str, name: Optional[str] = None) -> bool:
        """Download compute package from controller."""
        return self._package_runtime.download_compute_package(url, token, name)

    def set_compute_package_checksum(self, url: str, token: str, name: Optional[str] = None) -> bool:
        """Get checksum of compute package from controller."""
        return self._package_runtime.set_checksum(url, token, name)

    def unpack_compute_package(self) -> Tuple[bool, str]:
        """Unpack the compute package."""
        result, path = self._package_runtime.unpack_compute_package()
        if result:
            logger.info(f"Compute package unpacked to: {path}")
        else:
            logger.error("Error: Could not unpack compute package")

        return result, path

    def validate_compute_package(self, checksum: str) -> bool:
        """Validate the compute package."""
        return self._package_runtime.validate(checksum)

    def set_dispatcher(self, path: str) -> bool:
        """Set the dispatcher."""
        result = self._package_runtime.get_dispatcher(path)
        if result:
            self.dispatcher = result
            return True

        logger.error("Error: Could not set dispatcher")
        return False

    def get_or_set_environment(self) -> bool:
        """Get or set the environment."""
        try:
            logger.info("Initiating Dispatcher with entrypoint set to: startup")
            activate_cmd = self.dispatcher._get_or_create_python_env()
            self.dispatcher.run_cmd("startup")
        except KeyError:
            logger.info("No startup command found in package. Continuing.")
            return False
        except Exception as e:
            logger.error(f"Caught exception: {type(e).__name__}")
            return False

        if activate_cmd:
            logger.info(f"To activate the virtual environment, run: {activate_cmd}")

        return True

    def init_grpchandler(self, config: GrpcConnectionOptions, client_name: str, token: str) -> bool:
        """Initialize the GRPC handler."""
        try:
            if config.fqdn and len(config.fqdn) > 0:
                host = config.fqdn
                port = 443
            else:
                host = config.host
                port = config.port
            combiner_name = config.host

            self.grpc_handler = GrpcHandler(host=host, port=port, name=client_name, token=token, combiner_name=combiner_name)

            logger.info("Successfully initialized GRPC connection")
            return True
        except Exception as e:
            logger.error(f"Could not initialize GRPC connection: {e}")
            return False

    def send_heartbeats(self, client_name: str, client_id: str, update_frequency: float = 2.0) -> None:
        """Send heartbeats to the server."""
        self.grpc_handler.send_heartbeats(client_name=client_name, client_id=client_id, update_frequency=update_frequency)

    def listen_to_task_stream(self, client_name: str, client_id: str) -> None:
        """Listen to the task stream."""
        self.grpc_handler.listen_to_task_stream(client_name=client_name, client_id=client_id, callback=self._task_stream_callback)

    def _task_stream_callback(self, request: fedn.TaskRequest) -> None:
        """Handle task stream callbacks."""
        if request.type == fedn.StatusType.MODEL_UPDATE:
            self.update_local_model(request)
        elif request.type == fedn.StatusType.MODEL_VALIDATION:
            self.validate_global_model(request)
        elif request.type == fedn.StatusType.MODEL_PREDICTION:
            self.predict_global_model(request)

    def update_local_model(self, request: fedn.TaskRequest) -> None:
        """Update the local model."""
        model_id = request.model_id
        model_update_id = str(uuid.uuid4())

        tic = time.time()
        in_model = self.get_model_from_combiner(id=model_id, client_id=self.client_id)

        if in_model is None:
            logger.error("Could not retrieve model from combiner. Aborting training request.")
            return

        fetch_model_time = time.time() - tic

        if not self.train_callback:
            logger.error("No train callback set")
            return

        self.send_status(
            f"\t Starting processing of training request for model_id {model_id}",
            sesssion_id=request.session_id,
            sender_name=self.name,
            log_level=fedn.LogLevel.INFO,
            type=fedn.StatusType.MODEL_UPDATE,
        )

        logger.info(f"Running train callback with model ID: {model_id}")
        client_settings = json.loads(request.data).get("client_settings", {})
        tic = time.time()
        out_model, meta = self.train_callback(in_model, client_settings)
        meta["processing_time"] = time.time() - tic

        tic = time.time()
        self.send_model_to_combiner(model=out_model, id=model_update_id)
        meta["upload_model"] = time.time() - tic

        meta["fetch_model"] = fetch_model_time
        meta["config"] = request.data

        update = self.create_update_message(model_id=model_id, model_update_id=model_update_id, meta=meta, request=request)

        self.send_model_update(update)

        self.send_status(
            "Model update completed.",
            log_level=fedn.LogLevel.AUDIT,
            type=fedn.StatusType.MODEL_UPDATE,
            request=update,
            sesssion_id=request.session_id,
            sender_name=self.name,
        )

    def validate_global_model(self, request: fedn.TaskRequest) -> None:
        """Validate the global model."""
        model_id = request.model_id

        self.send_status(
            f"Processing validate request for model_id {model_id}",
            sesssion_id=request.session_id,
            sender_name=self.name,
            log_level=fedn.LogLevel.INFO,
            type=fedn.StatusType.MODEL_VALIDATION,
        )

        in_model = self.get_model_from_combiner(id=model_id, client_id=self.client_id)

        if in_model is None:
            logger.error("Could not retrieve model from combiner. Aborting validation request.")
            return

        if not self.validate_callback:
            logger.error("No validate callback set")
            return

        logger.info(f"Running validate callback with model ID: {model_id}")
        metrics = self.validate_callback(in_model)

        if metrics is not None:
            # Send validation
            validation = self.create_validation_message(metrics=metrics, request=request)

            result: bool = self.send_model_validation(validation)

            if result:
                self.send_status(
                    "Model validation completed.",
                    log_level=fedn.LogLevel.AUDIT,
                    type=fedn.StatusType.MODEL_VALIDATION,
                    request=validation,
                    sesssion_id=request.session_id,
                    sender_name=self.name,
                )
            else:
                self.send_status(
                    f"Client {self.name} failed to complete model validation.",
                    log_level=fedn.LogLevel.WARNING,
                    request=request,
                    sesssion_id=request.session_id,
                    sender_name=self.name,
                )

    def predict_global_model(self, request: fedn.TaskRequest) -> None:
        """Predict using the global model."""
        model_id = request.model_id
        model = self.get_model_from_combiner(id=model_id, client_id=self.client_id)

        if model is None:
            logger.error("Could not retrieve model from combiner. Aborting prediction request.")
            return

        if not self.predict_callback:
            logger.error("No predict callback set")
            return

        logger.info(f"Running predict callback with model ID: {model_id}")
        prediction = self.predict_callback(model)

        prediction_message = self.create_prediction_message(prediction=prediction, request=request)

        self.send_model_prediction(prediction_message)

    def create_update_message(self, model_id: str, model_update_id: str, meta: dict, request: fedn.TaskRequest) -> fedn.ModelUpdate:
        """Create an update message."""
        return self.grpc_handler.create_update_message(
            sender_name=self.name,
            model_id=model_id,
            model_update_id=model_update_id,
            receiver_name=request.sender.name,
            receiver_role=request.sender.role,
            meta=meta,
        )

    def create_validation_message(self, metrics: dict, request: fedn.TaskRequest) -> fedn.ModelValidation:
        """Create a validation message."""
        return self.grpc_handler.create_validation_message(
            sender_name=self.name,
            receiver_name=request.sender.name,
            receiver_role=request.sender.role,
            model_id=request.model_id,
            metrics=json.dumps(metrics),
            correlation_id=request.correlation_id,
            session_id=request.session_id,
        )

    def create_prediction_message(self, prediction: dict, request: fedn.TaskRequest) -> fedn.ModelPrediction:
        """Create a prediction message."""
        return self.grpc_handler.create_prediction_message(
            sender_name=self.name,
            receiver_name=request.sender.name,
            receiver_role=request.sender.role,
            model_id=request.model_id,
            prediction_output=json.dumps(prediction),
            correlation_id=request.correlation_id,
            session_id=request.session_id,
        )

    def set_name(self, name: str) -> None:
        """Set the client name."""
        logger.info(f"Setting client name to: {name}")
        self.name = name

    def set_client_id(self, client_id: str) -> None:
        """Set the client ID."""
        logger.info(f"Setting client ID to: {client_id}")
        self.client_id = client_id

    def run(self) -> None:
        """Run the client."""
        threading.Thread(target=self.send_heartbeats, kwargs={"client_name": self.name, "client_id": self.client_id}, daemon=True).start()
        try:
            self.listen_to_task_stream(client_name=self.name, client_id=self.client_id)
        except KeyboardInterrupt:
            logger.info("Client stopped by user.")

    def get_model_from_combiner(self, id: str, client_id: str, timeout: int = 20) -> BytesIO:
        """Get the model from the combiner."""
        return self.grpc_handler.get_model_from_combiner(id=id, client_id=client_id, timeout=timeout)

    def send_model_to_combiner(self, model: BytesIO, id: str) -> None:
        """Send the model to the combiner."""
        self.grpc_handler.send_model_to_combiner(model, id)

    def send_status(
        self,
        msg: str,
        log_level: fedn.LogLevel = fedn.LogLevel.INFO,
        type: Optional[str] = None,
        request: Optional[Union[fedn.ModelUpdate, fedn.ModelValidation, fedn.TaskRequest]] = None,
        sesssion_id: Optional[str] = None,
        sender_name: Optional[str] = None,
    ) -> None:
        """Send the status."""
        self.grpc_handler.send_status(msg, log_level, type, request, sesssion_id, sender_name)

    def send_model_update(self, update: fedn.ModelUpdate) -> bool:
        """Send the model update."""
        return self.grpc_handler.send_model_update(update)

    def send_model_validation(self, validation: fedn.ModelValidation) -> bool:
        """Send the model validation."""
        return self.grpc_handler.send_model_validation(validation)

    def send_model_prediction(self, prediction: fedn.ModelPrediction) -> bool:
        """Send the model prediction."""
        return self.grpc_handler.send_model_prediction(prediction)

    def init_remote_compute_package(self, url: str, token: str, package_checksum: Optional[str] = None) -> bool:
        """Initialize the remote compute package."""
        result = self.download_compute_package(url, token)
        if not result:
            logger.error("Could not download compute package")
            return False
        result = self.set_compute_package_checksum(url, token)
        if not result:
            logger.error("Could not set checksum")
            return False

        if package_checksum:
            result = self.validate_compute_package(package_checksum)
            if not result:
                logger.error("Could not validate compute package")
                return False

        result, path = self.unpack_compute_package()

        if not result:
            logger.error("Could not unpack compute package")
            return False

        logger.info(f"Compute package unpacked to: {path}")

        result = self.set_dispatcher(path)

        if not result:
            logger.error("Could not set dispatcher")
            return False

        logger.info("Dispatcher set")

        result = self.get_or_set_environment()

        return True

    def init_local_compute_package(self) -> bool:
        """Initialize the local compute package."""
        path = os.path.join(os.getcwd(), "client")
        result = self.set_dispatcher(path)

        if not result:
            logger.error("Could not set dispatcher")
            return False

        result = self.get_or_set_environment()

        logger.info("Dispatcher set")

        return True



"""Contains the PackageRuntime class, used to download, validate, and unpack compute packages."""

import cgi
import os
import tarfile
from typing import Optional, Tuple

import requests

from fedn.common.config import FEDN_AUTH_SCHEME
from fedn.common.log_config import logger
from fedn.utils.checksum import sha
from fedn.utils.dispatcher import Dispatcher, _read_yaml_file

# Constants for HTTP status codes
HTTP_STATUS_OK = 200
HTTP_STATUS_NO_CONTENT = 204

# Default timeout for requests
REQUEST_TIMEOUT = 10  # seconds


class PackageRuntime:
    """PackageRuntime is used to download, validate, and unpack compute packages.

    :param package_path: Path to compute package.
    :type package_path: str
    """

    def __init__(self, package_path: str) -> None:
        """Initialize the PackageRuntime."""
        self.dispatch_config = {
            "entry_points": {
                "predict": {"command": "python3 predict.py"},
                "train": {"command": "python3 train.py"},
                "validate": {"command": "python3 validate.py"},
            }
        }

        self.pkg_path = package_path
        self.pkg_name: Optional[str] = None
        self.checksum: Optional[str] = None

    def download_compute_package(self, url: str, token: str, name: Optional[str] = None) -> bool:
        """Download compute package from controller.

        :param url: URL of the controller.
        :param token: Token for authentication.
        :param name: Name of the package.
        :return: True if download was successful, False otherwise.
        :rtype: bool
        """
        try:
            path = f"{url}/api/v1/packages/download?name={name}" if name else f"{url}/api/v1/packages/download"

            with requests.get(path, stream=True, timeout=REQUEST_TIMEOUT, headers={"Authorization": f"{FEDN_AUTH_SCHEME} {token}"}) as r:
                if HTTP_STATUS_OK <= r.status_code < HTTP_STATUS_NO_CONTENT:
                    params = cgi.parse_header(r.headers.get("Content-Disposition", ""))[-1]
                    try:
                        self.pkg_name = params["filename"]
                    except KeyError:
                        logger.error("No package returned.")
                        return False
                    r.raise_for_status()
                    with open(os.path.join(self.pkg_path, self.pkg_name), "wb") as f:
                        for chunk in r.iter_content(chunk_size=8192):
                            f.write(chunk)

            return True
        except Exception:
            return False

    def set_checksum(self, url: str, token: str, name: Optional[str] = None) -> bool:
        """Get checksum of compute package from controller.

        :param url: URL of the controller.
        :param token: Token for authentication.
        :param name: Name of the package.
        :return: True if checksum was set successfully, False otherwise.
        :rtype: bool
        """
        try:
            path = f"{url}/api/v1/packages/checksum?name={name}" if name else f"{url}/api/v1/packages/checksum"

            with requests.get(path, timeout=REQUEST_TIMEOUT, headers={"Authorization": f"{FEDN_AUTH_SCHEME} {token}"}) as r:
                if HTTP_STATUS_OK <= r.status_code < HTTP_STATUS_NO_CONTENT:
                    data = r.json()
                    try:
                        self.checksum = data["checksum"]
                    except KeyError:
                        logger.error("Could not extract checksum.")

            return True
        except Exception:
            return False

    def validate(self, expected_checksum: str) -> bool:
        """Validate the package against the checksum provided by the controller.

        :param expected_checksum: Checksum provided by the controller.
        :return: True if checksums match, False otherwise.
        :rtype: bool
        """
        file_checksum = str(sha(os.path.join(self.pkg_path, self.pkg_name)))

        if self.checksum == expected_checksum == file_checksum:
            logger.info(f"Package validated {self.checksum}")
            return True
        return False

    def unpack_compute_package(self) -> Tuple[bool, str]:
        """Unpack the compute package.

        :return: Tuple containing a boolean indicating success and the path to the unpacked package.
        :rtype: Tuple[bool, str]
        """
        if not self.pkg_name:
            logger.error("Failed to unpack compute package, no pkg_name set. Has the reducer been configured with a compute package?")
            return False, ""

        try:
            if self.pkg_name.endswith(("tar.gz", ".tgz", "tar.bz2")):
                with tarfile.open(os.path.join(self.pkg_path, self.pkg_name), "r:*") as f:
                    for member in f.getmembers():
                        f.extract(member, self.pkg_path)
                    logger.info(f"Successfully extracted compute package content in {self.pkg_path}")
                    logger.info("Deleting temporary package tarball file.")
                    os.remove(os.path.join(self.pkg_path, self.pkg_name))

                    for root, _, files in os.walk(self.pkg_path):
                        if "fedn.yaml" in files:
                            logger.info(f"Found fedn.yaml file in {root}")
                            return True, root

                    logger.error("No fedn.yaml file found in extracted package!")
                    return False, ""
        except Exception:
            logger.error("Error extracting files.")
            os.remove(os.path.join(self.pkg_path, self.pkg_name))
            return False, ""

    def get_dispatcher(self, run_path: str) -> Optional[Dispatcher]:
        """Dispatch the compute package.

        :param run_path: Path to dispatch the compute package.
        :type run_path: str
        :return: Dispatcher object or None if an error occurred.
        :rtype: Optional[Dispatcher]
        """
        try:
            self.dispatch_config = _read_yaml_file(os.path.join(run_path, "fedn.yaml"))
            return Dispatcher(self.dispatch_config, run_path)
        except Exception:
            return None



from abc import abstractmethod
from typing import Any, Dict, List, Tuple

import pymongo
from bson import ObjectId
from pymongo.database import Database
from sqlalchemy import Integer, func, or_, select

from fedn.network.storage.statestore.stores.sql.shared import RoundCombinerModel, RoundConfigModel, RoundDataModel, RoundModel
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store

from .shared import from_document


class Round:
    def __init__(self, id: str, round_id: str, status: str, round_config: dict, combiners: List[dict], round_data: dict):
        self.id = id
        self.round_id = round_id
        self.status = status
        self.round_config = round_config
        self.combiners = combiners
        self.round_data = round_data


class RoundStore(Store[Round]):
    @abstractmethod
    def get_latest_round_id(self) -> int:
        pass


class MongoDBRoundStore(RoundStore, MongoDBStore[Round]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)
        self.database[self.collection].create_index([("round_id", pymongo.DESCENDING)])

    def get(self, id: str) -> Round:
        """Get an entity by id
        param id: The id of the entity
            type: str
        return: The entity
        """
        kwargs = {}
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            kwargs["_id"] = id_obj
        else:
            kwargs["round_id"] = id

        document = self.database[self.collection].find_one(kwargs)

        if document is None:
            return None

        return from_document(document)

    def update(self, id: str, item: Round) -> Tuple[bool, Any]:
        return super().update(id, item)

    def add(self, item: Round) -> Tuple[bool, Any]:
        round_id = item["round_id"]
        existing = self.database[self.collection].find_one({"round_id": round_id})

        if existing is not None:
            return False, "Round with round_id already exists"

        return super().add(item)

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for RoundStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs) -> Dict[int, List[Round]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
            description: The maximum number of entities to return
        param skip: The number of entities to skip
            type: int
            description: The number of entities to skip
        param sort_key: The key to sort by
            type: str
            description: The key to sort by
        param sort_order: The order to sort by
            type: pymongo.DESCENDING
            description: The order to sort by
        return: The entities
        """
        return super().list(limit, skip, sort_key or "round_id", sort_order, **kwargs)

    def get_latest_round_id(self) -> int:
        obj = self.database[self.collection].find_one(sort=[("_id", pymongo.DESCENDING)])
        if obj:
            return int(obj["round_id"])
        else:
            return 0


def from_row(row: RoundModel) -> Round:
    round_data = None
    if row.round_data is not None:
        round_data = {
            "time_commit": row.round_data.time_commit,
            "reduce": {
                "time_aggregate_model": row.round_data.reduce_time_aggregate_model,
                "time_fetch_model": row.round_data.reduce_time_fetch_model,
                "time_load_model": row.round_data.reduce_time_load_model,
            },
        }

    round_config = None

    if row.round_config is not None:
        round_config = {
            "aggregator": row.round_config.aggregator,
            "round_timeout": row.round_config.round_timeout,
            "buffer_size": row.round_config.buffer_size,
            "delete_models_storage": row.round_config.delete_models_storage,
            "clients_required": row.round_config.clients_required,
            "validate": row.round_config.validate,
            "helper_type": row.round_config.helper_type,
            "task": row.round_config.task,
            "model_id": row.round_config.model_id,
            "session_id": row.round_config.session_id,
            "round_id": row.round_config.round_id,
            "rounds": row.round_config.rounds,
        }

    combiners = [
        {
            "model_id": combiner.model_id,
            "name": combiner.name,
            "round_id": combiner.round_id,
            "status": combiner.status,
            "time_exec_training": combiner.time_exec_training,
            "config": {
                "_job_id": combiner.config__job_id,
                "aggregator": combiner.config_aggregator,
                "buffer_size": combiner.config_buffer_size,
                "clients_required": combiner.config_clients_required,
                "delete_models_storage": combiner.config_delete_models_storage,
                "helper_type": combiner.config_helper_type,
                "model_id": combiner.config_model_id,
                "round_id": combiner.config_round_id,
                "round_timeout": combiner.config_round_timeout,
                "rounds": combiner.config_rounds,
                "session_id": combiner.config_session_id,
                "task": combiner.config_task,
                "validate": combiner.config_validate,
            },
            "data": {
                "aggregation_time": {
                    "nr_aggregated_models": combiner.data_aggregation_time_nr_aggregated_models,
                    "time_model_aggregation": combiner.data_aggregation_time_time_model_aggregation,
                    "time_model_load": combiner.data_aggregation_time_time_model_load,
                },
                "nr_expected_updates": combiner.data_nr_expected_updates,
                "nr_required_updates": combiner.data_nr_required_updates,
                "time_combination": combiner.data_time_combination,
                "timeout": combiner.data_timeout,
            },
        }
        for combiner in row.combiners
    ]

    return {
        "id": row.id,
        "committed_at": row.committed_at,
        "round_id": row.round_id,
        "status": row.status,
        "round_config": round_config,
        "round_data": round_data,
        "combiners": combiners,
    }


class SQLRoundStore(RoundStore, SQLStore[Round]):
    def __init__(self, Session):
        super().__init__(Session)

    def get(self, id: str) -> Round:
        with self.Session() as session:
            stmt = select(RoundModel).where(or_(RoundModel.id == id, RoundModel.round_id == id))
            item = session.scalars(stmt).first()

            if item is None:
                return None

            return from_row(item)

    def update(self, id, item: Round) -> Tuple[bool, Any]:
        with self.Session() as session:
            stmt = select(RoundModel).where(or_(RoundModel.id == id, RoundModel.round_id == id))
            existing_item = session.scalars(stmt).first()

            if existing_item is None:
                return False, f"Entity with (id | round_id) {id} not found"

            if "round_data" in item and item["round_data"] is not None:
                round_data = item["round_data"]
                reduce = round_data["reduce"] if "reduce" in round_data else {}

                if existing_item.round_data is None:
                    round_data = RoundDataModel(
                        time_commit=round_data["time_commit"] if "time_commit" in round_data else None,
                        reduce_time_aggregate_model=reduce["time_aggregate_model"] if "time_aggregate_model" in reduce else None,
                        reduce_time_fetch_model=reduce["time_fetch_model"] if "time_fetch_model" in reduce else None,
                        reduce_time_load_model=reduce["time_load_model"] if "time_load_model" in reduce else None,
                    )
                    session.add(round_data)
                    existing_item.round_data = round_data
                else:
                    existing_item.round_data.time_commit = round_data["time_commit"] if "time_commit" in round_data else None
                    existing_item.round_data.reduce_time_aggregate_model = reduce["time_aggregate_model"] if "time_aggregate_model" in reduce else None
                    existing_item.round_data.reduce_time_fetch_model = reduce["time_fetch_model"] if "time_fetch_model" in reduce else None
                    existing_item.round_data.reduce_time_load_model = reduce["time_load_model"] if "time_load_model" in reduce else None

            if "round_config" in item and item["round_config"] is not None:
                if existing_item.round_config is None:
                    existing_item.round_config = RoundConfigModel(
                        aggregator=item["round_config"]["aggregator"],
                        round_timeout=item["round_config"]["round_timeout"],
                        buffer_size=item["round_config"]["buffer_size"],
                        delete_models_storage=item["round_config"]["delete_models_storage"],
                        clients_required=item["round_config"]["clients_required"],
                        validate=item["round_config"]["validate"],
                        helper_type=item["round_config"]["helper_type"],
                        task=item["round_config"]["task"],
                        model_id=item["round_config"]["model_id"],
                        session_id=item["round_config"]["session_id"],
                        round_id=item["round_config"]["round_id"],
                        rounds=item["round_config"]["rounds"],
                    )
                else:
                    existing_item.round_config.aggregator = item["round_config"]["aggregator"]
                    existing_item.round_config.round_timeout = item["round_config"]["round_timeout"]
                    existing_item.round_config.buffer_size = item["round_config"]["buffer_size"]
                    existing_item.round_config.delete_models_storage = item["round_config"]["delete_models_storage"]
                    existing_item.round_config.clients_required = item["round_config"]["clients_required"]
                    existing_item.round_config.validate = item["round_config"]["validate"]
                    existing_item.round_config.helper_type = item["round_config"]["helper_type"]
                    existing_item.round_config.task = item["round_config"]["task"]
                    existing_item.round_config.round_id = item["round_config"]["round_id"]
                    existing_item.round_config.rounds = item["round_config"]["rounds"]

                    if "model_id" in item["round_config"]:
                        existing_item.round_config.model_id = item["round_config"]["model_id"]

                    if "session_id" in item["round_config"]:
                        existing_item.round_config.session_id = item["round_config"]["session_id"]

            if "combiners" in item and item["combiners"] is not None:
                if existing_item.combiners is not None:
                    existing_item.combiners.clear()

                for combiner in item["combiners"]:
                    config = combiner["config"] if "config" in combiner else {}
                    data = combiner["data"] if "data" in combiner else {}
                    aggregation_time = data["aggregation_time"] if "aggregation_time" in data else {}

                    child = RoundCombinerModel(
                        model_id=combiner["model_id"],
                        name=combiner["name"],
                        round_id=combiner["round_id"],
                        status=combiner["status"],
                        time_exec_training=combiner["time_exec_training"],
                        config__job_id=config["_job_id"],
                        config_aggregator=config["aggregator"],
                        config_buffer_size=config["buffer_size"],
                        config_clients_required=config["clients_required"],
                        config_delete_models_storage=config["delete_models_storage"]
                        if isinstance(config["delete_models_storage"], bool)
                        else config["delete_models_storage"] == "True",
                        config_helper_type=config["helper_type"],
                        config_model_id=config["model_id"],
                        config_round_id=config["round_id"],
                        config_round_timeout=config["round_timeout"],
                        config_rounds=config["rounds"],
                        config_session_id=config["session_id"],
                        config_task=config["task"],
                        config_validate=config["validate"] if isinstance(config["validate"], bool) else config["validate"] == "True",
                        data_aggregation_time_nr_aggregated_models=aggregation_time["nr_aggregated_models"]
                        if "nr_aggregated_models" in aggregation_time
                        else None,
                        data_aggregation_time_time_model_aggregation=aggregation_time["time_model_aggregation"]
                        if "time_model_aggregation" in aggregation_time
                        else None,
                        data_aggregation_time_time_model_load=aggregation_time["time_model_load"] if "time_model_load" in aggregation_time else None,
                        data_nr_expected_updates=data["nr_expected_updates"] if "nr_expected_updates" in data else None,
                        data_nr_required_updates=data["nr_required_updates"] if "nr_required_updates" in data else None,
                        data_time_combination=data["time_combination"] if "time_combination" in data else None,
                        data_timeout=data["timeout"] if "timeout" in data else None,
                    )

                    existing_item.combiners.append(child)

            existing_item.status = item["status"]
            session.commit()

            return True, from_row(existing_item)

    def add(self, item: Round) -> Tuple[bool, Any]:
        with self.Session() as session:
            round_id = item["round_id"]
            stmt = select(RoundModel).where(RoundModel.round_id == round_id)
            existing_item = session.scalars(stmt).first()

            if existing_item is not None:
                return False, "Round with round_id already exists"

            entity = RoundModel(round_id=item["round_id"], status=item["status"])

            round_data: RoundDataModel = None
            if "round_data" in item:
                round_data = RoundDataModel(
                    time_commit=item["round_data"]["time_commit"],
                    reduce_time_aggregate_model=item["round_data"]["reduce"]["time_aggregate_model"],
                    reduce_time_fetch_model=item["round_data"]["reduce"]["time_fetch_model"],
                    reduce_time_load_model=item["round_data"]["reduce"]["time_load_model"],
                )
                entity.round_data = round_data

            round_config: RoundConfigModel = None

            if "round_config" in item:
                round_config = RoundConfigModel(
                    aggregator=item["round_config"]["aggregator"],
                    round_timeout=item["round_config"]["round_timeout"],
                    buffer_size=item["round_config"]["buffer_size"],
                    delete_models_storage=item["round_config"]["delete_models_storage"],
                    clients_required=item["round_config"]["clients_required"],
                    validate=item["round_config"]["validate"],
                    helper_type=item["round_config"]["helper_type"],
                    task=item["round_config"]["task"],
                    model_id=item["round_config"]["model_id"],
                    session_id=item["round_config"]["session_id"],
                    round_id=item["round_config"]["round_id"],
                    rounds=item["round_config"]["rounds"],
                )
                entity.round_config = round_config

            combiners: List[RoundCombinerModel] = []

            if "combiners" in item:
                combiners = []

                for combiner in item["combiners"]:
                    config = combiner["config"] if "config" in combiner else {}
                    data = combiner["data"] if "data" in combiner else {}
                    aggregation_time = data["aggregation_time"] if "aggregation_time" in data else {}

                    combiners.append(
                        RoundCombinerModel(
                            model_id=combiner["model_id"],
                            name=combiner["name"],
                            round_id=combiner["round_id"],
                            status=combiner["status"],
                            time_exec_training=combiner["time_exec_training"],
                            config__job_id=config["_job_id"],
                            config_aggregator=config["aggregator"],
                            config_buffer_size=config["buffer_size"],
                            config_clients_required=config["clients_required"],
                            config_delete_models_storage=config["delete_models_storage"]
                            if isinstance(config["delete_models_storage"], bool)
                            else config["delete_models_storage"] == "True",
                            config_helper_type=config["helper_type"],
                            config_model_id=config["model_id"],
                            config_round_id=config["round_id"],
                            config_round_timeout=config["round_timeout"],
                            config_rounds=config["rounds"],
                            config_session_id=config["session_id"],
                            config_task=config["task"],
                            config_validate=config["validate"] if isinstance(config["validate"], bool) else config["validate"] == "True",
                            data_aggregation_time_nr_aggregated_models=aggregation_time["nr_aggregated_models"]
                            if "nr_aggregated_models" in aggregation_time
                            else None,
                            data_aggregation_time_time_model_aggregation=aggregation_time["time_model_aggregation"]
                            if "time_model_aggregation" in aggregation_time
                            else None,
                            data_aggregation_time_time_model_load=aggregation_time["time_model_load"] if "time_model_load" in aggregation_time else None,
                            data_nr_expected_updates=data["nr_expected_updates"] if "nr_expected_updates" in data else None,
                            data_nr_required_updates=data["nr_required_updates"] if "nr_required_updates" in data else None,
                            data_time_combination=data["time_combination"] if "time_combination" in data else None,
                            data_timeout=data["timeout"] if "timeout" in data else None,
                        )
                    )

                entity.combiners = combiners

            session.add(entity)
            session.commit()

            return True, from_row(entity)

    def delete(self, id: str) -> bool:
        raise NotImplementedError

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs):
        with self.Session() as session:
            stmt = select(RoundModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(RoundModel, key) == value)

            if sort_key:
                _sort_order: str = "DESC" if sort_order == pymongo.DESCENDING else "ASC"
                _sort_key: str = sort_key or "committed_at"

                if _sort_key == "_id":
                    _sort_key = "id"

                if _sort_key == "round_id":
                    sort_obj = RoundModel.round_id.cast(Integer) if _sort_order == "ASC" else RoundModel.round_id.cast(Integer).desc()
                elif _sort_key in RoundModel.__table__.columns:
                    sort_obj = RoundModel.__table__.columns.get(_sort_key) if _sort_order == "ASC" else RoundModel.__table__.columns.get(_sort_key).desc()

                stmt = stmt.order_by(sort_obj)

            if limit:
                stmt = stmt.offset(skip or 0).limit(limit)
            if skip:
                stmt = stmt.offset(skip)

            items = session.execute(stmt)

            result = []

            for item in items:
                (r,) = item

                result.append(from_row(r))

            return {"count": len(result), "result": result}

    def count(self, **kwargs):
        with self.Session() as session:
            stmt = select(func.count()).select_from(RoundModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(RoundModel, key) == value)

            count = session.scalar(stmt)

            return count

    def get_latest_round_id(self) -> int:
        response = self.list(limit=1, skip=0, sort_key="round_id", sort_order=pymongo.DESCENDING)
        if response and "result" in response and len(response["result"]) > 0:
            round_id: str = response["result"][0]["round_id"]
            return int(round_id)
        else:
            return 0



from abc import abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Tuple

import pymongo
from bson import ObjectId
from pymongo.database import Database
from sqlalchemy import func, select
from sqlalchemy.orm import aliased

from fedn.network.storage.statestore.stores.shared import from_document
from fedn.network.storage.statestore.stores.sql.shared import ModelModel
from fedn.network.storage.statestore.stores.store import MongoDBStore, SQLStore, Store


class Model:
    def __init__(self, id: str, key: str, model: str, parent_model: str, session_id: str, committed_at: datetime):
        self.id = id
        self.key = key
        self.model = model
        self.parent_model = parent_model
        self.session_id = session_id
        self.committed_at = committed_at


class ModelStore(Store[Model]):
    @abstractmethod
    def list_descendants(self, id: str, limit: int) -> List[Model]:
        pass

    @abstractmethod
    def list_ancestors(self, id: str, limit: int, include_self: bool = False, reverse: bool = False) -> List[Model]:
        pass

    @abstractmethod
    def get_active(self) -> str:
        pass

    @abstractmethod
    def set_active(self, id: str) -> bool:
        pass


def validate(item: Model) -> Tuple[bool, str]:
    if "model" not in item or not item["model"]:
        return False, "Model is required"

    return True, ""


class MongoDBModelStore(ModelStore, MongoDBStore[Model]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)
        self.database[self.collection].create_index([("model", pymongo.DESCENDING)])

    def get(self, id: str) -> Model:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        return: The entity
        """
        kwargs = {"key": "models"}
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            kwargs["_id"] = id_obj
        else:
            kwargs["model"] = id

        document = self.database[self.collection].find_one(kwargs)

        if document is None:
            return None

        return from_document(document)

    def _complement(self, item: Model):
        if "key" not in item or item["key"] is None:
            item["key"] = "models"

    def update(self, id: str, item: Model) -> Tuple[bool, Any]:
        valid, message = validate(item)
        if not valid:
            return False, message

        self._complement(item)

        return super().update(id, item)

    def add(self, item: Model) -> Tuple[bool, Any]:
        valid, message = validate(item)
        if not valid:
            return False, message

        self._complement(item)

        return super().add(item)

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for ModelStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs) -> Dict[int, List[Model]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        kwargs["key"] = "models"

        return super().list(limit, skip, sort_key or "committed_at", sort_order, **kwargs)

    def list_descendants(self, id: str, limit: int) -> List[Model]:
        """List descendants
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        param limit: The maximum number of entities to return
            type: int
        return: A list of entities
        """
        kwargs = {"key": "models"}
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            kwargs["_id"] = id_obj
        else:
            kwargs["model"] = id

        model: object = self.database[self.collection].find_one(kwargs)

        if model is None:
            return None

        current_model_id: str = model["model"]
        result: list = []

        for _ in range(limit):
            if current_model_id is None:
                break

            model: str = self.database[self.collection].find_one({"key": "models", "parent_model": current_model_id})

            if model is not None:
                formatted_model = Model.from_dict(model)
                result.append(formatted_model)
                current_model_id = model["model"]
            else:
                break

        result.reverse()

        return result

    def list_ancestors(self, id: str, limit: int, include_self: bool = False, reverse: bool = False) -> List[Model]:
        """List ancestors
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        param limit: The maximum number of entities to return
            type: int
        return: A list of entities
        """
        kwargs = {"key": "models"}
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            kwargs["_id"] = id_obj
        else:
            kwargs["model"] = id

        model: object = self.database[self.collection].find_one(kwargs)

        if model is None:
            return None

        current_model_id: str = model["parent_model"]
        result: list = []

        if include_self:
            formatted_model = from_document(model)
            result.append(formatted_model)

        for _ in range(limit):
            if current_model_id is None:
                break

            model = self.database[self.collection].find_one({"key": "models", "model": current_model_id})

            if model is not None:
                formatted_model = from_document(model)
                result.append(formatted_model)
                current_model_id = model["parent_model"]
            else:
                break

        if reverse:
            result.reverse()

        return result

    def count(self, **kwargs) -> int:
        """Count entities
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: The count (int)
        """
        kwargs["key"] = "models"
        return super().count(**kwargs)

    def get_active(self) -> str:
        """Get the active model
        return: The active model id (str)
        """
        active_model = self.database[self.collection].find_one({"key": "current_model"})

        if active_model is None:
            return None

        return active_model["model"]

    def set_active(self, id: str) -> bool:
        """Set the active model
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        return: True if successful
        """
        kwargs = {"key": "models"}
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            kwargs["_id"] = id_obj
        else:
            kwargs["model"] = id

        model = self.database[self.collection].find_one(kwargs)

        if model is None:
            return False

        self.database[self.collection].update_one({"key": "current_model"}, {"$set": {"model": model["model"]}}, upsert=True)

        return True


def from_row(row: ModelModel) -> Model:
    return {
        "id": row.id,
        "model": row.id,
        "committed_at": row.committed_at,
        "parent_model": row.parent_model,
        "session_id": row.session_id,
        "name": row.name,
        "active": row.active,
    }


class SQLModelStore(ModelStore, SQLStore[Model]):
    def __init__(self, Session):
        super().__init__(Session)

    def get(self, id: str) -> Model:
        with self.Session() as session:
            stmt = select(ModelModel).where(ModelModel.id == id)
            item = session.scalars(stmt).first()
            if item is None:
                return None
            return from_row(item)

    def update(self, id: str, item: Model) -> Tuple[bool, Any]:
        valid, message = validate(item)
        if not valid:
            return False, message
        with self.Session() as session:
            stmt = select(ModelModel).where(ModelModel.id == id)
            existing_item = session.execute(stmt).first()
            if existing_item is None:
                return False, "Item not found"

            existing_item.parent_model = item["parent_model"]
            existing_item.name = item["name"]
            existing_item.session_id = item.get("session_id")
            existing_item.committed_at = item["committed_at"]
            existing_item.active = item["active"]

            return True, from_row(existing_item)

    def add(self, item: Model) -> Tuple[bool, Any]:
        valid, message = validate(item)
        if not valid:
            return False, message

        with self.Session() as session:
            id: str = None
            if "model" in item:
                id = item["model"]
            elif "id" in item:
                id = item["id"]

            item = ModelModel(
                id=id,
                parent_model=item["parent_model"],
                name=item["name"],
                session_id=item.get("session_id"),
                committed_at=item["committed_at"],
                active=item["active"] if "active" in item else False,
            )
            session.add(item)
            session.commit()
            return True, from_row(item)

    def delete(self, id: str) -> bool:
        raise NotImplementedError

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs):
        with self.Session() as session:
            stmt = select(ModelModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(ModelModel, key) == value)

            _sort_order: str = "DESC" if sort_order == pymongo.DESCENDING else "ASC"
            _sort_key: str = sort_key or "committed_at"

            if _sort_key in ModelModel.__table__.columns:
                sort_obj = ModelModel.__table__.columns.get(_sort_key) if _sort_order == "ASC" else ModelModel.__table__.columns.get(_sort_key).desc()

                stmt = stmt.order_by(sort_obj)

            if limit:
                stmt = stmt.offset(skip or 0).limit(limit)
            elif skip:
                stmt = stmt.offset(skip)

            items = session.scalars(stmt).all()

            result = []
            for i in items:
                result.append(from_row(i))

            count = session.scalar(select(func.count()).select_from(ModelModel))

            return {"count": count, "result": result}

    def count(self, **kwargs):
        with self.Session() as session:
            stmt = select(func.count()).select_from(ModelModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(ModelModel, key) == value)

            count = session.scalar(stmt)

            return count

    def list_descendants(self, id: str, limit: int):
        with self.Session() as session:
            # Define the recursive CTE
            descendant = aliased(ModelModel)  # Alias for recursion
            cte = select(ModelModel).where(ModelModel.parent_model == id).cte(name="descendant_cte", recursive=True)
            cte = cte.union_all(select(descendant).where(descendant.parent_model == cte.c.id))

            # Final query with optional limit
            query = select(cte)
            if limit is not None:
                query = query.limit(limit)

            # Execute the query
            items = session.execute(query).fetchall()

            # Return the list of descendants
            result = []
            for i in items:
                result.append(from_row(i))

            return result

    def list_ancestors(self, id: str, limit: int, include_self=False, reverse=False):
        with self.Session() as session:
            # Define the recursive CTE
            ancestor = aliased(ModelModel)  # Alias for recursion
            cte = select(ModelModel).where(ModelModel.id == id).cte(name="ancestor_cte", recursive=True)
            cte = cte.union_all(select(ancestor).where(ancestor.id == cte.c.parent_model))

            # Final query with optional limit
            query = select(cte)
            if limit is not None:
                query = query.limit(limit)

            # Execute the query
            items = session.execute(query).fetchall()

            # Return the list of ancestors
            result = []
            for i in items:
                result.append(from_row(i))

            return result

    def get_active(self) -> str:
        with self.Session() as session:
            active_stmt = select(ModelModel).where(ModelModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                return active_item.id
            return None

    def set_active(self, id: str) -> bool:
        with self.Session() as session:
            active_stmt = select(ModelModel).where(ModelModel.active)
            active_item = session.scalars(active_stmt).first()
            if active_item:
                active_item.active = False

            stmt = select(ModelModel).where(ModelModel.id == id)
            item = session.scalars(stmt).first()

            if item is None:
                return False

            item.active = True
            session.commit()
        return True



import json
import queue
import re
import signal
import threading
import time
import uuid
from datetime import datetime, timedelta
from enum import Enum
from typing import TypedDict

import pymongo
from google.protobuf.json_format import MessageToDict

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.certificate.certificate import Certificate
from fedn.common.log_config import logger, set_log_level_from_string, set_log_stream
from fedn.network.combiner.roundhandler import RoundConfig, RoundHandler
from fedn.network.combiner.shared import client_store, combiner_store, prediction_store, repository, round_store, status_store, validation_store
from fedn.network.grpc.server import Server, ServerConfig

VALID_NAME_REGEX = "^[a-zA-Z0-9_-]*$"


class Role(Enum):
    """Enum for combiner roles."""

    WORKER = 1
    COMBINER = 2
    REDUCER = 3
    OTHER = 4


def role_to_proto_role(role):
    """Convert a Role to a proto Role.

    :param role: the role to convert
    :type role: :class:`fedn.network.combiner.server.Role`
    :return: proto role
    :rtype: :class:`fedn.network.grpc.fedn_pb2.Role`
    """
    if role == Role.COMBINER:
        return fedn.COMBINER
    if role == Role.WORKER:
        return fedn.CLIENT
    if role == Role.REDUCER:
        return fedn.REDUCER
    if role == Role.OTHER:
        return fedn.OTHER


class CombinerConfig(TypedDict):
    """Configuration for the combiner."""

    discover_host: str
    discover_port: int
    token: str
    host: str
    port: int
    ip: str
    parent: str
    fqdn: str
    name: str
    secure: bool
    verify: bool
    cert_path: str
    key_path: str
    max_clients: int
    network_id: str
    logfile: str
    verbosity: str


# TODO: dependency injection
class Combiner(rpc.CombinerServicer, rpc.ReducerServicer, rpc.ConnectorServicer, rpc.ControlServicer):
    """Combiner gRPC server.

    :param config: configuration for the combiner
    :type config: dict
    """

    def __init__(self, config):
        """Initialize Combiner server."""
        set_log_level_from_string(config.get("verbosity", "INFO"))
        set_log_stream(config.get("logfile", None))

        # Client queues
        # Each client in the dict is stored with its client_id as key, and the value is a dict with keys:
        #     name: str
        #     status: str
        #     last_seen: str
        #     fedn.Queue.TASK_QUEUE: queue.Queue
        # Obs that fedn.Queue.TASK_QUEUE is just str(1)
        self.clients = {}

        # Validate combiner name
        match = re.search(VALID_NAME_REGEX, config["name"])
        if not match:
            raise ValueError("Unallowed character in combiner name. Allowed characters: a-z, A-Z, 0-9, _, -.")

        self.id = config["name"]
        self.role = Role.COMBINER
        self.max_clients = config["max_clients"]

        # Set up model repository
        self.repository = repository

        self.round_store = round_store

        # Add combiner to statestore
        interface_config = {
            "port": config["port"],
            "fqdn": config["fqdn"],
            "name": config["name"],
            "address": config["host"],
            "parent": "localhost",
            "ip": "",
            "updated_at": str(datetime.now()),
        }
        # Check if combiner already exists in statestore
        if combiner_store.get(config["name"]) is None:
            combiner_store.add(interface_config)

        # Fetch all clients previously connected to the combiner
        # If a client and a combiner goes down at the same time,
        # the client will be stuck listed as "online" in the statestore.
        # Set the status to offline for previous clients.
        previous_clients = client_store.list(limit=0, skip=0, sort_key=None, sort_order=pymongo.DESCENDING, **{"combiner": self.id})
        count = previous_clients["count"]
        result = previous_clients["result"]
        logger.info(f"Found {count} previous clients")
        logger.info("Updating previous clients status to offline")
        for client in result:
            try:
                client_to_update = client_store.get(client["client_id"])
                client_to_update["status"] = "offline"
                client_store.update(client["client_id"], client_to_update)

            except Exception as e:
                logger.error("Failed to update previous client status: {}".format(str(e)))

        # Set up gRPC server configuration
        if config["secure"]:
            cert = Certificate(key_path=config["key_path"], cert_path=config["cert_path"])
            certificate, key = cert.get_keypair_raw()
            grpc_server_config = ServerConfig(port=config["port"], secure=True, key=key, certificate=certificate)
        else:
            grpc_server_config = ServerConfig(port=config["port"], secure=False)

        # Create gRPC server
        self.server = Server(self, grpc_server_config)

        # Set up round controller
        self.round_handler = RoundHandler(self)

        # Start thread for round controller
        threading.Thread(target=self.round_handler.run, daemon=True).start()

        # Start thread for client status updates: TODO: Should be configurable
        threading.Thread(target=self._deamon_thread_client_status, daemon=True).start()

        # Start the gRPC server
        self.server.start()

    def __whoami(self, client, instance):
        """Set the client id and role in a proto message.

        :param client: the client to set the id and role for
        :type client: :class:`fedn.network.grpc.fedn_pb2.Client`
        :param instance: the instance to get the id and role from
        :type instance: :class:`fedn.network.combiner.server.Combiner`
        :return: the client with id and role set
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Client`
        """
        client.name = instance.id
        client.role = role_to_proto_role(instance.role)
        return client

    def request_model_update(self, session_id, model_id, config, clients=[]):
        """Ask clients to update the current global model.

        :param config: the model configuration to send to clients
        :type config: dict
        :param clients: the clients to send the request to
        :type clients: list

        """
        clients = self._send_request_type(fedn.StatusType.MODEL_UPDATE, session_id, model_id, config, clients)

        if len(clients) < 20:
            logger.info("Sent model update request for model {} to clients {}".format(model_id, clients))
        else:
            logger.info("Sent model update request for model {} to {} clients".format(model_id, len(clients)))

    def request_model_validation(self, session_id, model_id, clients=[]):
        """Ask clients to validate the current global model.

        :param model_id: the model id to validate
        :type model_id: str
        :param config: the model configuration to send to clients
        :type config: dict
        :param clients: the clients to send the request to
        :type clients: list

        """
        clients = self._send_request_type(fedn.StatusType.MODEL_VALIDATION, session_id, model_id, clients)

        if len(clients) < 20:
            logger.info("Sent model validation request for model {} to clients {}".format(model_id, clients))
        else:
            logger.info("Sent model validation request for model {} to {} clients".format(model_id, len(clients)))

    def request_model_prediction(self, prediction_id: str, model_id: str, clients: list = []) -> None:
        """Ask clients to perform prediction on the model.

        :param model_id: the model id to perform prediction on
        :type model_id: str
        :param config: the model configuration to send to clients
        :type config: dict
        :param clients: the clients to send the request to
        :type clients: list

        """
        clients = self._send_request_type(fedn.StatusType.MODEL_PREDICTION, prediction_id, model_id, clients)

        if len(clients) < 20:
            logger.info("Sent model prediction request for model {} to clients {}".format(model_id, clients))
        else:
            logger.info("Sent model prediction request for model {} to {} clients".format(model_id, len(clients)))

    def _send_request_type(self, request_type, session_id, model_id, config=None, clients=[]):
        """Send a request of a specific type to clients.

        :param request_type: the type of request
        :type request_type: :class:`fedn.network.grpc.fedn_pb2.StatusType`
        :param session_id: the session id to send in the request. Obs that for prediction, this is the prediction id.
        :type session_id: str
        :param model_id: the model id to send in the request
        :type model_id: str
        :param config: the model configuration to send to clients
        :type config: dict
        :param clients: the clients to send the request to
        :type clients: list
        :return: the clients
        :rtype: list
        """
        if len(clients) == 0:
            if request_type == fedn.StatusType.MODEL_UPDATE:
                clients = self.get_active_trainers()
            elif request_type == fedn.StatusType.MODEL_VALIDATION:
                clients = self.get_active_validators()
            elif request_type == fedn.StatusType.MODEL_PREDICTION:
                # TODO: add prediction clients type
                clients = self.get_active_validators()
        for client in clients:
            request = fedn.TaskRequest()
            request.model_id = model_id
            request.correlation_id = str(uuid.uuid4())
            request.timestamp = str(datetime.now())
            request.type = request_type
            request.session_id = session_id

            request.sender.name = self.id
            request.sender.role = fedn.COMBINER
            request.receiver.client_id = client
            request.receiver.role = fedn.CLIENT
            # Set the request data, not used in validation
            if request_type == fedn.StatusType.MODEL_PREDICTION:
                presigned_url = self.repository.presigned_put_url(self.repository.prediction_bucket, f"{client}/{session_id}")
                # TODO: in prediction, request.data should also contain user-defined data/parameters
                request.data = json.dumps({"presigned_url": presigned_url})
            elif request_type == fedn.StatusType.MODEL_UPDATE:
                request.data = json.dumps(config)
            self._put_request_to_client_queue(request, fedn.Queue.TASK_QUEUE)
        return clients

    def get_active_trainers(self):
        """Get a list of active trainers.

        :return: the list of active trainers
        :rtype: list
        """
        trainers = self._list_active_clients(fedn.Queue.TASK_QUEUE)
        return trainers

    def get_active_validators(self):
        """Get a list of active validators.

        :return: the list of active validators
        :rtype: list
        """
        validators = self._list_active_clients(fedn.Queue.TASK_QUEUE)
        return validators

    def nr_active_trainers(self):
        """Get the number of active trainers.

        :return: the number of active trainers
        :rtype: int
        """
        return len(self.get_active_trainers())

    ####################################################################################################################

    def __join_client(self, client):
        """Add a client to the list of active clients.

        :param client: the client to add
        :type client: :class:`fedn.network.grpc.fedn_pb2.Client`
        """
        if client.client_id not in self.clients.keys():
            # The status is set to offline by default, and will be updated once _list_active_clients is called.
            self.clients[client.client_id] = {"last_seen": datetime.now(), "status": "offline"}

    def _subscribe_client_to_queue(self, client, queue_name):
        """Subscribe a client to the queue.

        :param client: the client to subscribe
        :type client: :class:`fedn.network.grpc.fedn_pb2.Client`
        :param queue_name: the name of the queue to subscribe to
        :type queue_name: str
        """
        self.__join_client(client)
        if queue_name not in self.clients[client.client_id].keys():
            self.clients[client.client_id][queue_name] = queue.Queue()

    def __get_queue(self, client, queue_name):
        """Get the queue for a client.

        :param client: the client to get the queue for
        :type client: :class:`fedn.network.grpc.fedn_pb2.Client`
        :param queue_name: the name of the queue to get
        :type queue_name: str
        :return: the queue
        :rtype: :class:`queue.Queue`

        :raises KeyError: if the queue does not exist
        """
        try:
            return self.clients[client.client_id][queue_name]
        except KeyError:
            raise

    def _list_subscribed_clients(self, queue_name):
        """List all clients subscribed to a queue.

        :param queue_name: the name of the queue
        :type queue_name: str
        :return: a list of client names
        :rtype: list
        """
        subscribed_clients = []
        for name, client in self.clients.items():
            if queue_name in client.keys():
                subscribed_clients.append(name)
        return subscribed_clients

    def _list_active_clients(self, channel):
        """List all clients that have sent a status message in the last 10 seconds.

        :param channel: the name of the channel
        :type channel: str
        :return: a list of client names
        :rtype: list
        """
        # Temporary dict to store client status
        clients = {
            "active_clients": [],
            "update_active_clients": [],
            "update_offline_clients": [],
        }
        for client in self._list_subscribed_clients(channel):
            status = self.clients[client]["status"]
            now = datetime.now()
            then = self.clients[client]["last_seen"]
            if (now - then) < timedelta(seconds=10):
                clients["active_clients"].append(client)
                # If client has changed status, update client queue
                if status != "online":
                    self.clients[client]["status"] = "online"
                    clients["update_active_clients"].append(client)
            elif status != "offline":
                self.clients[client]["status"] = "offline"
                clients["update_offline_clients"].append(client)
        # Update statestore with client status
        if len(clients["update_active_clients"]) > 0:
            for client in clients["update_active_clients"]:
                client_to_update = client_store.get(client)
                client_to_update["status"] = "online"
                client_store.update(client, client_to_update)
        if len(clients["update_offline_clients"]) > 0:
            for client in clients["update_offline_clients"]:
                client_to_update = client_store.get(client)
                client_to_update["status"] = "offline"
                client_store.update(client, client_to_update)

        return clients["active_clients"]

    def _deamon_thread_client_status(self, timeout=5):
        """Deamon thread that checks for inactive clients and updates statestore."""
        while True:
            time.sleep(timeout)
            # TODO: Also update validation clients
            self._list_active_clients(fedn.Queue.TASK_QUEUE)

    def _put_request_to_client_queue(self, request, queue_name):
        """Get a client specific queue and add a request to it.
        The client is identified by the request.receiver.

        :param request: the request to send
        :type request: :class:`fedn.network.grpc.fedn_pb2.Request`
        :param queue_name: the name of the queue to send the request to
        :type queue_name: str
        """
        try:
            q = self.__get_queue(request.receiver, queue_name)
            q.put(request)
        except Exception as e:
            logger.error("Failed to put request to client queue {} for client {}: {}".format(queue_name, request.receiver.name, str(e)))
            raise

    def _send_status(self, status):
        """Report a status to backend db.

        :param status: the status message to report
        :type status: :class:`fedn.network.grpc.fedn_pb2.Status`
        """
        data = MessageToDict(status)
        _ = status_store.add(data)

    def _flush_model_update_queue(self):
        """Clear the model update queue (aggregator).

        :return: True if successful, else False
        """
        q = self.round_handler.aggregator.model_updates
        try:
            with q.mutex:
                q.queue.clear()
                q.all_tasks_done.notify_all()
                q.unfinished_tasks = 0
            return True
        except Exception as e:
            logger.error("Failed to flush model update queue: %s", str(e))
            return False

    #####################################################################################################################

    # Controller Service

    def Start(self, control: fedn.ControlRequest, context):
        """Start a round of federated learning"

        :param control: the control request
        :type control: :class:`fedn.network.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        logger.info("grpc.Combiner.Start: Starting round")

        config = RoundConfig()
        for parameter in control.parameter:
            config.update({parameter.key: parameter.value})

        logger.debug("grpc.Combiner.Start: Round config {}".format(config))

        job_id = self.round_handler.push_round_config(config)
        logger.info("grcp.Combiner.Start: Pushed round config (job_id): {}".format(job_id))

        response = fedn.ControlResponse()
        p = response.parameter.add()
        p.key = "job_id"
        p.value = job_id

        return response

    def SetAggregator(self, control: fedn.ControlRequest, context):
        """Set the active aggregator.

        :param control: the control request
        :type control: :class:`fedn.network.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        logger.debug("grpc.Combiner.SetAggregator: Called")
        for parameter in control.parameter:
            aggregator = parameter.value
        status = self.round_handler.set_aggregator(aggregator)

        response = fedn.ControlResponse()
        if status:
            response.message = "Success"
        else:
            response.message = "Failed"
        return response

    def SetServerFunctions(self, control: fedn.ControlRequest, context):
        """Set a function provider.

        :param control: the control request
        :type control: :class:`fedn.network.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        logger.debug("grpc.Combiner.SetServerFunctions: Called")
        for parameter in control.parameter:
            server_functions = parameter.value

        self.round_handler.set_server_functions(server_functions)

        response = fedn.ControlResponse()
        response.message = "Success"
        logger.info(f"set function provider response {response}")
        return response

    def FlushAggregationQueue(self, control: fedn.ControlRequest, context):
        """Flush the queue.

        :param control: the control request
        :type control: :class:`fedn.network.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        logger.debug("grpc.Combiner.FlushAggregationQueue: Called")
        status = self._flush_model_update_queue()

        response = fedn.ControlResponse()
        if status:
            response.message = "Success"
        else:
            response.message = "Failed"

        return response

    ##############################################################################

    def Stop(self, control: fedn.ControlRequest, context):
        """TODO: Not yet implemented.

        :param control: the control request
        :type control: :class:`fedn.network.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        response = fedn.ControlResponse()
        logger.info("grpc.Combiner.Stop: Called")
        return response

    #####################################################################################################################

    def SendStatus(self, status: fedn.Status, context):
        """A client RPC endpoint that accepts status messages.

        :param status: the status message
        :type status: :class:`fedn.network.grpc.fedn_pb2.Status`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Response`
        """
        logger.debug("grpc.Combiner.SendStatus: Called")
        self._send_status(status)

        response = fedn.Response()
        response.response = "Status received."
        return response

    def ListActiveClients(self, request: fedn.ListClientsRequest, context):
        """RPC endpoint that returns a ClientList containing the names of all active clients.
            An active client has sent a status message / responded to a heartbeat
            request in the last 10 seconds.

        :param request: the request
        :type request: :class:`fedn.network.grpc.fedn_pb2.ListClientsRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the client list
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ClientList`
        """
        clients = fedn.ClientList()
        active_clients = self._list_active_clients(request.channel)
        nr_active_clients = len(active_clients)
        if nr_active_clients < 20:
            logger.info("grpc.Combiner.ListActiveClients:  Active clients: {}".format(active_clients))
        else:
            logger.info("grpc.Combiner.ListActiveClients: Number active clients: {}".format(nr_active_clients))

        for client in active_clients:
            clients.client.append(fedn.Client(name=client, role=fedn.CLIENT))
        return clients

    def AcceptingClients(self, request: fedn.ConnectionRequest, context):
        """RPC endpoint that returns a ConnectionResponse indicating whether the server
        is accepting clients or not.

        :param request: the request (unused)
        :type request: :class:`fedn.network.grpc.fedn_pb2.ConnectionRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ConnectionResponse`
        """
        response = fedn.ConnectionResponse()
        active_clients = self._list_active_clients(fedn.Queue.TASK_QUEUE)

        try:
            requested = int(self.max_clients)
            if len(active_clients) >= requested:
                response.status = fedn.ConnectionStatus.NOT_ACCEPTING
                return response
            if len(active_clients) < requested:
                response.status = fedn.ConnectionStatus.ACCEPTING
                return response

        except Exception as e:
            logger.error("Combiner not properly configured! {}".format(e))
            raise

        response.status = fedn.ConnectionStatus.TRY_AGAIN_LATER
        return response

    def SendHeartbeat(self, heartbeat: fedn.Heartbeat, context):
        """RPC that lets clients send a hearbeat, notifying the server that
            the client is available.

        :param heartbeat: the heartbeat
        :type heartbeat: :class:`fedn.network.grpc.fedn_pb2.Heartbeat`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Response`
        """
        logger.debug("GRPC: Received heartbeat from {}".format(heartbeat.sender.name))
        # Update the clients dict with the last seen timestamp.
        client = heartbeat.sender
        self.__join_client(client)
        self.clients[client.client_id]["last_seen"] = datetime.now()

        response = fedn.Response()
        response.sender.name = heartbeat.sender.name
        response.sender.role = heartbeat.sender.role
        response.response = "Heartbeat received"
        return response

    # Combiner Service

    def TaskStream(self, response, context):
        """A server stream RPC endpoint (Update model). Messages from client stream.

        :param response: the response
        :type response: :class:`fedn.network.grpc.fedn_pb2.ModelUpdateRequest`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        """
        client = response.sender
        metadata = context.invocation_metadata()
        if metadata:
            metadata = dict(metadata)
            logger.info("grpc.Combiner.TaskStream: Client connected: {}\n".format(metadata["client"]))

        status = fedn.Status(status="Client {} connecting to TaskStream.".format(client.name), log_level=fedn.LogLevel.INFO, type=fedn.StatusType.NETWORK)
        logger.info("Client {} connecting to TaskStream.".format(client.name))
        status.timestamp.GetCurrentTime()

        self.__whoami(status.sender, self)

        # Subscribe client, this also adds the client to self.clients
        self._subscribe_client_to_queue(client, fedn.Queue.TASK_QUEUE)
        q = self.__get_queue(client, fedn.Queue.TASK_QUEUE)

        self._send_status(status)

        # Set client status to online
        self.clients[client.client_id]["status"] = "online"
        try:
            # If the client is already in the client store, update the status

            client_to_upsert = {
                "name": client.name,
                "status": "online",
                "client_id": client.client_id,
                "last_seen": datetime.now(),
                "combiner": self.id,
                "updated_at": datetime.now(),
            }

            success, result = client_store.upsert(client_to_upsert)
            if not success:
                logger.error(result)
        except Exception as e:
            logger.error(f"Failed to update client status: {str(e)}")

        # Keep track of the time context has been active
        start_time = time.time()
        while context.is_active():
            # Check if the context has been active for more than 10 seconds
            if time.time() - start_time > 10:
                self.clients[client.client_id]["last_seen"] = datetime.now()
                # Reset the start time
                start_time = time.time()
            try:
                yield q.get(timeout=1.0)
            except queue.Empty:
                pass
            except Exception as e:
                logger.error("Error in ModelUpdateRequestStream: {}".format(e))
        logger.warning("Client {} disconnected from TaskStream".format(client.name))
        status = fedn.Status(status="Client {} disconnected from TaskStream.".format(client.name))
        status.log_level = fedn.LogLevel.INFO
        status.type = fedn.StatusType.NETWORK
        status.timestamp.GetCurrentTime()
        self.__whoami(status.sender, self)
        self._send_status(status)

    def SendModelUpdate(self, request, context):
        """Send a model update response.

        :param request: the request
        :type request: :class:`fedn.network.grpc.fedn_pb2.ModelUpdate`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Response`
        """
        self.round_handler.update_handler.on_model_update(request)

        response = fedn.Response()
        response.response = "RECEIVED ModelUpdate {} from client  {}".format(response, response.sender.name)
        return response  # TODO Fill later

    def register_model_validation(self, validation):
        """Register a model validation.

        :param validation: the model validation
        :type validation: :class:`fedn.network.grpc.fedn_pb2.ModelValidation`
        """
        data = MessageToDict(validation)
        success, result = validation_store.add(data)
        if not success:
            logger.error(result)
        else:
            logger.info("Model validation registered: {}".format(result))

    def SendModelValidation(self, request, context):
        """Send a model validation response.

        :param request: the request
        :type request: :class:`fedn.network.grpc.fedn_pb2.ModelValidation`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Response`
        """
        logger.info("Recieved ModelValidation from {}".format(request.sender.name))

        validation = MessageToDict(request)
        validation_store.add(validation)

        response = fedn.Response()
        response.response = "RECEIVED ModelValidation {} from client  {}".format(response, response.sender.name)
        return response

    def SendModelPrediction(self, request, context):
        """Send a model prediction response.

        :param request: the request
        :type request: :class:`fedn.network.grpc.fedn_pb2.ModelPrediction`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Response`
        """
        logger.info("Recieved ModelPrediction from {}".format(request.sender.name))

        result = MessageToDict(request)
        prediction_store.add(result)

        response = fedn.Response()
        response.response = "RECEIVED ModelPrediction {} from client  {}".format(response, response.sender.name)
        return response

    ####################################################################################################################

    def run(self):
        """Start the server."""
        logger.info("COMBINER: {} started, ready for gRPC requests.".format(self.id))
        try:
            while True:
                signal.pause()
        except (KeyboardInterrupt, SystemExit):
            pass
        self.server.stop()
        self.server.stop()



from concurrent import futures
from typing import TypedDict

import grpc
from grpc_health.v1 import health, health_pb2_grpc

import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.log_config import logger, set_log_level_from_string, set_log_stream
from fedn.network.combiner.shared import modelservice
from fedn.network.grpc.auth import JWTInterceptor


class ServerConfig(TypedDict):
    port: int
    secure: bool
    key: str
    certificate: str
    logfile: str
    verbosity: str


class Server:
    """Class for configuring and launching the gRPC server."""

    def __init__(self, servicer, config: ServerConfig):
        set_log_level_from_string(config.get("verbosity", "INFO"))
        set_log_stream(config.get("logfile", None))

        # Keepalive settings: these detect if the client is alive
        KEEPALIVE_TIME_MS = 60 * 1000  # send keepalive ping every 60 seconds
        KEEPALIVE_TIMEOUT_MS = 20 * 1000  # wait 20 seconds for keepalive ping ack before considering connection dead
        MAX_CONNECTION_IDLE_MS = 5 * 60 * 1000  # max idle time before server terminates the connection (5 minutes)

        self.server = grpc.server(
            futures.ThreadPoolExecutor(max_workers=350),
            interceptors=[JWTInterceptor()],
            options=[
                ("grpc.keepalive_time_ms", KEEPALIVE_TIME_MS),
                ("grpc.keepalive_timeout_ms", KEEPALIVE_TIMEOUT_MS),
                ("grpc.max_connection_idle_ms", MAX_CONNECTION_IDLE_MS),
            ],
        )
        self.certificate = None
        self.health_servicer = health.HealthServicer()

        if isinstance(servicer, rpc.CombinerServicer):
            rpc.add_CombinerServicer_to_server(servicer, self.server)
        if isinstance(servicer, rpc.ConnectorServicer):
            rpc.add_ConnectorServicer_to_server(servicer, self.server)
        if isinstance(servicer, rpc.ReducerServicer):
            rpc.add_ReducerServicer_to_server(servicer, self.server)
        if isinstance(modelservice, rpc.ModelServiceServicer):
            rpc.add_ModelServiceServicer_to_server(modelservice, self.server)
        if isinstance(servicer, rpc.CombinerServicer):
            rpc.add_ControlServicer_to_server(servicer, self.server)

        health_pb2_grpc.add_HealthServicer_to_server(self.health_servicer, self.server)

        if config["secure"]:
            logger.info("Creating secure gRPCS server using certificate")
            server_credentials = grpc.ssl_server_credentials(
                (
                    (
                        config["key"],
                        config["certificate"],
                    ),
                )
            )
            self.server.add_secure_port("[::]:" + str(config["port"]), server_credentials)
        else:
            logger.info("Creating gRPC server")
            self.server.add_insecure_port("[::]:" + str(config["port"]))

    def start(self):
        """Start the gRPC server."""
        logger.info("gRPC Server started")
        self.server.start()

    def stop(self):
        """Stop the gRPC server."""
        logger.info("gRPC Server stopped")
        self.server.stop(0)



import pytest
from unittest.mock import patch
import os

from fedn.network.storage.dbconnection import DatabaseConnection
from fedn.tests.stores.helpers.mongo_docker import start_mongodb_container, stop_mongodb_container
from fedn.tests.stores.helpers.postgres_docker import start_postgres_container, stop_postgres_container



def network_id():
    return "test_network"

@pytest.fixture(scope="package")
def mongo_connection():
    if not os.environ.get("UNITTEST_GITHUB"):
        _, port = start_mongodb_container()
    else: 
        port = 27017

    def mongo_config():
        return {
            "type": "MongoDB",
            "mongo_config": {
                "host": "localhost",
                "port": port,
                "username": os.environ.get("UNITTEST_DBUSER", "_"),
                "password": os.environ.get("UNITTEST_DBPASS", "_"),
            }
        }

    with patch('fedn.network.storage.dbconnection.get_statestore_config', return_value=mongo_config()), \
         patch('fedn.network.storage.dbconnection.get_network_config', return_value=network_id()):
        yield DatabaseConnection(force_create_new=True)

    if not os.environ.get("UNITTEST_GITHUB"):
        stop_mongodb_container()
    
@pytest.fixture(scope="package")
def sql_connection():
    def sql_config():
        return {
            "type": "SQLite",
            "sqlite_config": {
                "dbname": ":memory:",
            }
        }

    with patch('fedn.network.storage.dbconnection.get_statestore_config', return_value=sql_config()), \
         patch('fedn.network.storage.dbconnection.get_network_config', return_value=network_id()):
        return DatabaseConnection(force_create_new=True)
    
@pytest.fixture(scope="package")
def postgres_connection():
    if not os.environ.get("UNITTEST_GITHUB"):
        _, port = start_postgres_container()
    else:
        port = 5432


    def postgres_config():
        return {
            "type": "PostgreSQL",
            "postgres_config": {
                "username": os.environ.get("UNITTEST_DBUSER", "_"),
                "password": os.environ.get("UNITTEST_DBPASS", "_"),
                "host": "localhost",
                "port": port
            }
        }

    with patch('fedn.network.storage.dbconnection.get_statestore_config', return_value=postgres_config()), \
         patch('fedn.network.storage.dbconnection.get_network_config', return_value=network_id()):
        yield DatabaseConnection(force_create_new=True)
    if not os.environ.get("UNITTEST_GITHUB"):
        stop_postgres_container()


import json
from concurrent import futures

import grpc

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.log_config import logger
from fedn.network.combiner.hooks.allowed_import import *  # noqa: F403

# imports for user code
from fedn.network.combiner.hooks.allowed_import import ServerFunctionsBase
from fedn.network.combiner.modelservice import bytesIO_request_generator, model_as_bytesIO, unpack_model
from fedn.utils.helpers.plugins.numpyhelper import Helper

CHUNK_SIZE = 1024 * 1024
VALID_NAME_REGEX = "^[a-zA-Z0-9_-]*$"


class FunctionServiceServicer(rpc.FunctionServiceServicer):
    """Function service running in an environment combined with each combiner.

    Receiving requests from the combiner.
    """

    def __init__(self) -> None:
        """Initialize long-running Function server."""
        super().__init__()

        self.helper = Helper()
        self.server_functions: ServerFunctionsBase = None
        self.server_functions_code: str = None
        self.client_updates = {}
        self.implemented_functions = {}

    def HandleClientConfig(self, request_iterator: fedn.ClientConfigRequest, context):
        """Distribute client configs to clients from user defined code.

        :param request_iterator: the client config request
        :type request_iterator: :class:`fedn.network.grpc.fedn_pb2.ClientConfigRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the client config response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ClientConfigResponse`
        """
        logger.info("Received client config request.")
        model, _ = unpack_model(request_iterator, self.helper)
        client_settings = self.server_functions.client_settings(global_model=model)
        logger.info(f"Client config response: {client_settings}")
        return fedn.ClientConfigResponse(client_settings=json.dumps(client_settings))

    def HandleClientSelection(self, request: fedn.ClientSelectionRequest, context):
        """Handle client selection from user defined code.

        :param request: the client selection request
        :type request: :class:`fedn.network.grpc.fedn_pb2.fedn.ClientSelectionRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the client selection response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ClientSelectionResponse`
        """
        logger.info("Received client selection request.")
        client_ids = json.loads(request.client_ids)
        client_ids = self.server_functions.client_selection(client_ids)
        logger.info(f"Clients selected: {client_ids}")
        return fedn.ClientSelectionResponse(client_ids=json.dumps(client_ids))

    def HandleMetadata(self, request: fedn.ClientMetaRequest, context):
        """Store client metadata from a request.

        :param request: the client meta request
        :type request: :class:`fedn.network.grpc.fedn_pb2.fedn.ClientMetaRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the client meta response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ClientMetaResponse`
        """
        logger.info("Received metadata")
        client_id = request.client_id
        metadata = json.loads(request.metadata)
        self.client_updates[client_id] = self.client_updates.get(client_id, []) + [metadata]
        return fedn.ClientMetaResponse(status="Metadata stored")

    def HandleStoreModel(self, request_iterator, context):
        model, final_request = unpack_model(request_iterator, self.helper)
        client_id = final_request.id
        if client_id == "global_model":
            logger.info("Received previous global model")
            self.previous_global = model
        else:
            logger.info("Received client model")
            self.client_updates[client_id] = [model] + self.client_updates.get(client_id, [])
        return fedn.StoreModelResponse(status=f"Received model originating from {client_id}")

    def HandleAggregation(self, request, context):
        """Receive and store models and aggregate based on user-defined code when specified in the request.

        :param request_iterator: the aggregation request
        :type request_iterator: :class:`fedn.network.grpc.fedn_pb2.fedn.AggregationRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the aggregation response (aggregated model or None)
        :rtype: :class:`fedn.network.grpc.fedn_pb2.AggregationResponse`
        """
        logger.info(f"Receieved aggregation request: {request.aggregate}")
        aggregated_model = self.server_functions.aggregate(self.previous_global, self.client_updates)
        model_bytesIO = model_as_bytesIO(aggregated_model, self.helper)
        request_function = fedn.AggregationResponse
        self.client_updates = {}
        logger.info("Returning aggregate model.")
        response_generator = bytesIO_request_generator(mdl=model_bytesIO, request_function=request_function, args={})
        for response in response_generator:
            yield response

    def HandleProvidedFunctions(self, request: fedn.ProvidedFunctionsResponse, context):
        """Handles the 'provided_functions' request. Sends back which functions are available.

        :param request: the provided function request
        :type request: :class:`fedn.network.grpc.fedn_pb2.fedn.ProvidedFunctionsRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: dict with str -> bool for which functions are available
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ProvidedFunctionsResponse`
        """
        logger.info("Receieved provided functions request.")
        server_functions_code = request.function_code
        # if no new code return previous
        if server_functions_code == self.server_functions_code:
            logger.info("No new server function code provided.")
            logger.info(f"Provided function: {self.implemented_functions}")
            return fedn.ProvidedFunctionsResponse(available_functions=self.implemented_functions)

        self.server_functions_code = server_functions_code
        self.implemented_functions = {}
        self._instansiate_server_functions_code()
        # We are not sending dummy values here since the implementation might depend on model shape / implementations

        # Implemented=False if return is None (indicating base implementation)

        # check if aggregation is available
        try:
            ret = self.server_functions.aggregate(0, 0)
            if ret is None:
                self.implemented_functions["aggregate"] = False
            else:
                self.implemented_functions["aggregate"] = True
        except Exception:
            self.implemented_functions["aggregate"] = True
        # check if client_settings is available
        try:
            ret = self.server_functions.client_settings(0)
            if ret is None:
                self.implemented_functions["client_settings"] = False
            else:
                self.implemented_functions["client_settings"] = True
        except Exception:
            self.implemented_functions["client_settings"] = True
        # check if client_selection is available
        try:
            ret = self.server_functions.client_selection(0)
            if ret is None:
                self.implemented_functions["client_selection"] = False
            else:
                self.implemented_functions["client_selection"] = True
        except Exception:
            self.implemented_functions["client_selection"] = True
        logger.info(f"Provided function: {self.implemented_functions}")
        return fedn.ProvidedFunctionsResponse(available_functions=self.implemented_functions)

    def _instansiate_server_functions_code(self):
        # this will create a new user defined instance of the ServerFunctions class.
        try:
            namespace = {}
            exec(self.server_functions_code, globals(), namespace)  # noqa: S102
            exec("server_functions = ServerFunctions()", globals(), namespace)  # noqa: S102
            self.server_functions = namespace.get("server_functions")
        except Exception as e:
            logger.error(f"Exec failed with error: {str(e)}")


def serve():
    """Start the hooks service."""
    # Keepalive settings: these detect if the client is alive
    KEEPALIVE_TIME_MS = 60 * 1000  # send keepalive ping every 60 seconds
    KEEPALIVE_TIMEOUT_MS = 20 * 1000  # wait 20 seconds for keepalive ping ack before considering connection dead
    MAX_CONNECTION_IDLE_MS = 5 * 60 * 1000  # max idle time before server terminates the connection (5 minutes)
    MAX_MESSAGE_LENGTH = 1 * 1024 * 1024 * 1024  # 1 GB in bytes
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=100),  # Increase based on expected load
        options=[
            ("grpc.keepalive_time_ms", KEEPALIVE_TIME_MS),
            ("grpc.keepalive_timeout_ms", KEEPALIVE_TIMEOUT_MS),
            ("grpc.max_connection_idle_ms", MAX_CONNECTION_IDLE_MS),
            ("grpc.max_send_message_length", MAX_MESSAGE_LENGTH),
            ("grpc.max_receive_message_length", -1),
        ],
    )
    rpc.add_FunctionServiceServicer_to_server(FunctionServiceServicer(), server)
    server.add_insecure_port("[::]:12081")
    server.start()
    server.wait_for_termination()



import random  # noqa: F401
from typing import Dict, List, Tuple  # noqa: F401

import numpy as np  # noqa: F401

from fedn.common.log_config import logger  # noqa: F401
from fedn.network.combiner.hooks.serverfunctionsbase import ServerFunctionsBase  # noqa: F401

print = logger.info



from abc import abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

import pymongo
from bson import ObjectId
from pymongo.database import Database
from sqlalchemy import String, func, or_, select
from sqlalchemy.orm import Mapped, mapped_column

from fedn.network.storage.statestore.stores.store import MongoDBStore, MyAbstractBase, SQLStore, Store

from .shared import from_document


class Client:
    def __init__(self, id: str, name: str, combiner: str, combiner_preferred: str, ip: str, status: str, updated_at: str, last_seen: datetime):
        self.id = id
        self.name = name
        self.combiner = combiner
        self.combiner_preferred = combiner_preferred
        self.ip = ip
        self.status = status
        self.updated_at = updated_at
        self.last_seen = last_seen


class ClientStore(Store[Client]):
    @abstractmethod
    def upsert(self, item: Client) -> Tuple[bool, Any]:
        pass

    @abstractmethod
    def connected_client_count(self, combiners: List[str]) -> List[Client]:
        pass


class MongoDBClientStore(ClientStore, MongoDBStore[Client]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)
        self.database[self.collection].create_index([("client_id", pymongo.DESCENDING)])

    def get(self, id: str) -> Client:
        """Get an entity by id
        param id: The id of the entity
            type: str
        return: The entity
        """
        if ObjectId.is_valid(id):
            response = super().get(id)
        else:
            obj = self._get_client_by_client_id(id)
            response = from_document(obj)
        return response

    def _get_client_by_client_id(self, client_id: str) -> Dict:
        document = self.database[self.collection].find_one({"client_id": client_id})
        if document is None:
            return None
        return document

    def update(self, id: str, item: Client) -> Tuple[bool, Any]:
        try:
            existing_client = self.get(id)

            return super().update(existing_client["id"], item)
        except Exception as e:
            return False, str(e)

    def add(self, item: Client) -> Tuple[bool, Any]:
        return super().add(item)

    def upsert(self, item: Client) -> Tuple[bool, Any]:
        try:
            result = self.database[self.collection].update_one(
                {"client_id": item["client_id"]}, {"$set": {k: v for k, v in item.items() if v is not None}}, upsert=True
            )
            id = result.upserted_id
            document = self.database[self.collection].find_one({"_id": id})
            return True, from_document(document)
        except Exception as e:
            return False, str(e)

    def delete(self, id: str) -> bool:
        kwargs = {"_id": ObjectId(id)} if ObjectId.is_valid(id) else {"client_id": id}

        document = self.database[self.collection].find_one(kwargs)

        if document is None:
            return False

        return super().delete(document["_id"])

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs) -> Dict[int, List[Client]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        return super().list(limit, skip, sort_key or "last_seen", sort_order, **kwargs)

    def count(self, **kwargs) -> int:
        return super().count(**kwargs)

    def connected_client_count(self, combiners: List[str]) -> List[Client]:
        """Count the number of connected clients for each combiner.

        :param combiners: list of combiners to get data for.
        :type combiners: list
        :param sort_key: The key to sort by.
        :type sort_key: str
        :param sort_order: The sort order.
        :type sort_order: pymongo.ASCENDING or pymongo.DESCENDING
        :return: list of combiner data.
        :rtype: list(ObjectId)
        """
        try:
            pipeline = (
                [
                    {"$match": {"combiner": {"$in": combiners}, "status": "online"}},
                    {"$group": {"_id": "$combiner", "count": {"$sum": 1}}},
                    {"$project": {"id": "$_id", "count": 1, "_id": 0}},
                ]
                if len(combiners) > 0
                else [
                    {"$match": {"status": "online"}},
                    {"$group": {"_id": "$combiner", "count": {"$sum": 1}}},
                    {"$project": {"id": "$_id", "count": 1, "_id": 0}},
                ]
            )

            result = list(self.database[self.collection].aggregate(pipeline))
        except Exception:
            result = {}

        return result


class ClientModel(MyAbstractBase):
    __tablename__ = "clients"

    client_id: Mapped[str] = mapped_column(String(255), unique=True)
    combiner: Mapped[str] = mapped_column(String(255))
    ip: Mapped[Optional[str]] = mapped_column(String(255))
    name: Mapped[str] = mapped_column(String(255))
    package: Mapped[Optional[str]] = mapped_column(String(255))
    status: Mapped[str] = mapped_column(String(255))
    last_seen: Mapped[datetime] = mapped_column(default=datetime.now())


def from_row(row: ClientModel) -> Client:
    return {
        "id": row.id,
        "client_id": row.client_id,
        "combiner": row.combiner,
        "ip": row.ip,
        "name": row.name,
        "package": row.package,
        "status": row.status,
        "last_seen": row.last_seen,
    }


class SQLClientStore(ClientStore, SQLStore[Client]):
    def __init__(self, Session):
        super().__init__(Session)

    def get(self, id: str) -> Client:
        with self.Session() as session:
            stmt = select(ClientModel).where(or_(ClientModel.id == id, ClientModel.client_id == id))
            item = session.scalars(stmt).first()

            if item is None:
                return None

            return from_row(item)

    def update(self, id: str, item: Client) -> Tuple[bool, Any]:
        with self.Session() as session:
            stmt = select(ClientModel).where(or_(ClientModel.id == id, ClientModel.client_id == id))
            existing_item = session.scalars(stmt).first()

            if existing_item is None:
                return False, "Item not found"

            existing_item.combiner = item.get("combiner")
            existing_item.ip = item.get("ip")
            existing_item.name = item.get("name")
            existing_item.package = item.get("package")
            existing_item.status = item.get("status")
            existing_item.last_seen = item.get("last_seen")

            session.commit()

            return True, from_row(existing_item)

    def add(self, item: Client) -> Tuple[bool, Any]:
        with self.Session() as session:
            entity = ClientModel(
                client_id=item.get("client_id"),
                combiner=item.get("combiner"),
                ip=item.get("ip"),
                name=item.get("name"),
                package=item.get("package"),
                status=item.get("status"),
                last_seen=item.get("last_seen"),
            )

            session.add(entity)
            session.commit()

            return True, from_row(entity)

    def delete(self, id):
        raise NotImplementedError

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, **kwargs):
        with self.Session() as session:
            stmt = select(ClientModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(ClientModel, key) == value)

            _sort_order: str = "DESC" if sort_order == pymongo.DESCENDING else "ASC"
            _sort_key: str = sort_key or "committed_at"

            if _sort_key in ClientModel.__table__.columns:
                sort_obj = ClientModel.__table__.columns.get(_sort_key) if _sort_order == "ASC" else ClientModel.__table__.columns.get(_sort_key).desc()

                stmt = stmt.order_by(sort_obj)

            if limit:
                stmt = stmt.offset(skip or 0).limit(limit)
            elif skip:
                stmt = stmt.offset(skip)

            items = session.scalars(stmt).all()

            result = []
            for i in items:
                result.append(from_row(i))

            count = session.scalar(select(func.count()).select_from(ClientModel))

            return {"count": count, "result": result}

    def count(self, **kwargs):
        with self.Session() as session:
            stmt = select(func.count()).select_from(ClientModel)

            for key, value in kwargs.items():
                stmt = stmt.where(getattr(ClientModel, key) == value)

            count = session.scalar(stmt)

            return count

    def upsert(self, item: Client) -> Tuple[bool, Any]:
        with self.Session() as session:
            id = item.get("id")
            client_id = item.get("client_id")

            stmt = select(ClientModel).where(or_(ClientModel.id == id, ClientModel.client_id == client_id))
            existing_item = session.scalars(stmt).first()

            if existing_item is None:
                entity = ClientModel(
                    client_id=item.get("client_id"),
                    combiner=item.get("combiner"),
                    ip=item.get("ip"),
                    name=item.get("name"),
                    package=item.get("package"),
                    status=item.get("status"),
                    last_seen=item.get("last_seen"),
                )

                session.add(entity)
                session.commit()

                return True, from_row(entity)

            existing_item.combiner = item.get("combiner")
            existing_item.ip = item.get("ip")
            existing_item.name = item.get("name")
            existing_item.package = item.get("package")
            existing_item.status = item.get("status")
            existing_item.last_seen = item.get("last_seen")

            session.commit()

            return True, from_row(existing_item)

    def connected_client_count(self, combiners):
        with self.Session() as session:
            stmt = select(ClientModel.combiner, func.count(ClientModel.combiner)).group_by(ClientModel.combiner)
            if combiners:
                stmt = stmt.where(ClientModel.combiner.in_(combiners))

            items = session.execute(stmt).fetchall()

            result = []
            for i in items:
                result.append({"combiner": i[0], "count": i[1]})

            return result



import pytest
import pymongo

import itertools
import datetime
import uuid

from fedn.network.storage.dbconnection import DatabaseConnection


@pytest.fixture
def test_clients():
    #TODO: SQL versions does not support updated_at field
    # TODO: Create using Client class
    start_date = datetime.datetime(2021, 1, 4, 1, 2, 4)
    return [{"client_id":str(uuid.uuid4()), "name":"test_client1", "combiner":"test_combiner", "ip":"121.12.32.2", "combiner_preferred":"", "status":"test_status", 
                   "last_seen":start_date - datetime.timedelta(days=1), "package":"local"},
            {"client_id":str(uuid.uuid4()), "name":"test_client2", "combiner":"test_combiner", "ip":"121.12.32.22", "combiner_preferred":"", "status":"test_status2", 
                   "last_seen":start_date - datetime.timedelta(days=3), "package":"local"},
            {"client_id":str(uuid.uuid4()), "name":"test_client3", "combiner":"test_combiner", "ip":"121.12.32.22", "combiner_preferred":"", "status":"test_status2", 
                   "last_seen":start_date - datetime.timedelta(days=4), "package":"local"},
            {"client_id":str(uuid.uuid4()), "name":"test_client5", "combiner":"test_combiner", "ip":"121.12.32.22", "combiner_preferred":"", "status":"test_status2", 
                   "last_seen":start_date - datetime.timedelta(seconds=5), "package":"remote"},
            {"client_id":str(uuid.uuid4()), "name":"test_client4", "combiner":"test_combiner2", "ip":"121.12.32.22", "combiner_preferred":"", "status":"test_status2", 
                   "last_seen":start_date - datetime.timedelta(weeks=1), "package":"remote"},
            {"client_id":str(uuid.uuid4()), "name":"test_client6", "combiner":"test_combiner2", "ip":"121.12.32.22", "combiner_preferred":"", "status":"test_status2", 
                   "last_seen":start_date - datetime.timedelta(hours=1), "package":"local"},
            {"client_id":str(uuid.uuid4()), "name":"test_client7", "combiner":"test_combiner2", "ip":"121.12.32.22", "combiner_preferred":"", "status":"test_status2", 
                   "last_seen":start_date - datetime.timedelta(minutes=1), "package":"remote"},
                   ]

@pytest.fixture
def db_connections_with_data(postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection, test_clients):
    for c in test_clients:
        res, _ = mongo_connection.client_store.add(c)
        assert res == True

    for c in test_clients:
        res, _ = postgres_connection.client_store.add(c)
        assert res == True

    for c in test_clients:
        res, _ = sql_connection.client_store.add(c)
        assert res == True

    yield [("postgres", postgres_connection), ("sqlite", sql_connection), ("mongo", mongo_connection)]

    # TODO:Clean up



@pytest.fixture
def options():
    sorting_keys = (#None, 
                        "name",
                        "client_id",
                        "last_seen",
                        #"invalid_key"
                        ) 
    limits = (None, 0, 1, 2, 99)
    skips = (None, 0, 1, 2, 99)
    desc = (None, pymongo.DESCENDING, pymongo.ASCENDING)
    opt_kwargs = ({}, {"ip":"121.12.32.22"}, {"combiner":"test_combiner", "status":"test_status"}, {"name":"test_client1", "status":"test_status2"})

    return list(itertools.product(limits, skips, sorting_keys, desc, opt_kwargs))


class TestClientStore:

    def test_add_update_delete(self, postgres_connection:DatabaseConnection, sql_connection: DatabaseConnection, mongo_connection:DatabaseConnection):
        pass

    def test_list(self, db_connections_with_data: list[tuple[str, DatabaseConnection]], options: list[tuple]):   
        for (name1, db_1), (name2, db_2) in zip(db_connections_with_data[1:], db_connections_with_data[:-1]):
            print("Running tests between databases {} and {}".format(name1, name2))
            for *opt,kwargs in options:                        
                res = db_1.client_store.list(*opt, **kwargs)
                count2, gathered_clients = res["count"], res["result"]

                res = db_2.client_store.list(*opt, **kwargs)
                count, gathered_clients2 = res["count"], res["result"]
                #TODO: The count is not equal to the number of clients in the list, but the number of clients returned by the query before skip and limit
                #It is not clear what is the intended behavior
                # assert(count == len(gathered_clients))
                # assert count == count2
                assert len(gathered_clients) == len(gathered_clients2)

                for i in range(len(gathered_clients)):
                    assert gathered_clients[i]["name"] == gathered_clients2[i]["name"]
                





    
    
    
    
    
